from pydantic import FilePath

from ..baserecordingextractorinterface import BaseRecordingExtractorInterface
from ....tools import get_package
from ....utils import DeepDict


class EDFRecordingInterface(BaseRecordingExtractorInterface):
    """
    Data interface class for converting European Data Format (EDF) data.

    Uses the :py:func:`~spikeinterface.extractors.read_edf` reader from SpikeInterface.

    Not supported on M1 macs.
    """

    display_name = "EDF Recording"
    keywords = BaseRecordingExtractorInterface.keywords + ("European Data Format",)
    associated_suffixes = (".edf",)
    info = "Interface for European Data Format (EDF) recording data."

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = super().get_source_schema()
        source_schema["properties"]["file_path"]["description"] = "Path to the .edf file."
        return source_schema

    @staticmethod
    def get_available_channel_ids(file_path: FilePath) -> list:
        """
        Get all available channel names from an EDF file.

        Parameters
        ----------
        file_path : FilePath
            Path to the EDF file

        Returns
        -------
        list
            List of all channel names in the EDF file
        """
        from spikeinterface.extractors import read_edf

        # Load the recording to inspect channels
        recording = read_edf(file_path=file_path, all_annotations=True, use_names_as_ids=True)

        # Get all channel IDs
        channel_ids = recording.get_channel_ids()

        # Clean up to avoid dangling references
        del recording

        return channel_ids.tolist()

    def _source_data_to_extractor_kwargs(self, source_data: dict) -> dict:

        extractor_kwargs = source_data.copy()
        extractor_kwargs.pop("channels_to_skip")
        extractor_kwargs["all_annotations"] = True
        extractor_kwargs["use_names_as_ids"] = True

        return extractor_kwargs

    def __init__(
        self,
        file_path: FilePath,
        verbose: bool = False,
        es_key: str = "ElectricalSeries",
        channels_to_skip: list | None = None,
    ):
        """
        Load and prepare data for EDF.
        Currently, only continuous EDF+ files (EDF+C) and original EDF files (EDF) are supported


        Parameters
        ----------
        file_path : str or Path
            Path to the edf file
        verbose : bool, default: False
            Allows verbose.
        es_key : str, default: "ElectricalSeries"
            Key for the ElectricalSeries metadata
        channels_to_skip : list, default: None
            Channels to skip when adding the data to the nwbfile. These parameter can be used to skip non-neural
            channels that are present in the EDF file.

        """
        get_package(
            package_name="pyedflib",
            excluded_platforms_and_python_versions=dict(darwin=dict(arm=["3.9"])),
        )

        super().__init__(file_path=file_path, verbose=verbose, es_key=es_key, channels_to_skip=channels_to_skip)
        self.edf_header = self.recording_extractor.neo_reader.edf_header

        # We remove the channels that are not neural
        if channels_to_skip:
            self.recording_extractor = self.recording_extractor.remove_channels(remove_channel_ids=channels_to_skip)

    def extract_nwb_file_metadata(self) -> dict:
        nwbfile_metadata = dict(
            session_start_time=self.edf_header["startdate"],
            experimenter=self.edf_header["technician"],
        )

        # Filter empty values
        nwbfile_metadata = {property: value for property, value in nwbfile_metadata.items() if value}

        return nwbfile_metadata

    def extract_subject_metadata(self) -> dict:
        subject_metadata = dict(
            subject_id=self.edf_header["patientcode"],
            date_of_birth=self.edf_header["birthdate"],
        )

        # Filter empty values
        subject_metadata = {property: value for property, value in subject_metadata.items() if value}

        return subject_metadata

    def get_metadata(self) -> DeepDict:
        metadata = super().get_metadata()
        nwbfile_metadata = self.extract_nwb_file_metadata()
        metadata["NWBFile"].update(nwbfile_metadata)

        subject_metadata = self.extract_subject_metadata()
        metadata.get("Subject", dict()).update(subject_metadata)

        return metadata



"""General purpose iterator for all ImagingExtractor data."""

import numpy as np
from roiextractors import ImagingExtractor
from tqdm import tqdm

from neuroconv.tools.hdmf import GenericDataChunkIterator
from neuroconv.tools.iterative_write import (
    get_image_series_buffer_shape,
    get_image_series_chunk_shape,
)


class ImagingExtractorDataChunkIterator(GenericDataChunkIterator):
    """DataChunkIterator for ImagingExtractor objects primarily used when writing imaging data to an NWB file."""

    def __init__(
        self,
        imaging_extractor: ImagingExtractor,
        buffer_gb: float | None = None,
        buffer_shape: tuple | None = None,
        chunk_mb: float | None = None,
        chunk_shape: tuple | None = None,
        display_progress: bool = False,
        progress_bar_class: tqdm | None = None,
        progress_bar_options: dict | None = None,
    ):
        """
        Initialize an Iterable object which returns DataChunks with data and their selections on each iteration.

        Parameters
        ----------
        imaging_extractor : ImagingExtractor
            The ImagingExtractor object which handles the data access.
        buffer_gb : float, optional
            The upper bound on size in gigabytes (GB) of each selection from the iteration.
            The buffer_shape will be set implicitly by this argument.
            Cannot be set if `buffer_shape` is also specified.
            The default is 1GB.
        buffer_shape : tuple, optional
            Manual specification of buffer shape to return on each iteration.
            Must be a multiple of chunk_shape along each axis.
            Cannot be set if `buffer_gb` is also specified.
            The default is None.
        chunk_mb : float, optional
            The upper bound on size in megabytes (MB) of the internal chunk for the HDF5 dataset.
            The chunk_shape will be set implicitly by this argument.
            Cannot be set if `chunk_shape` is also specified.
            The default is 10MB, as recommended by the HDF5 group.
            For more details, search the hdf5 documentation for "Improving IO Performance Compressed Datasets".
        chunk_shape : tuple, optional
            Manual specification of the internal chunk shape for the HDF5 dataset.
            Cannot be set if `chunk_mb` is also specified.
            The default is None.
        display_progress : bool, default=False
            Display a progress bar with iteration rate and estimated completion time.
        progress_bar_class : dict, optional
            The progress bar class to use.
            Defaults to tqdm.tqdm if the TQDM package is installed.
        progress_bar_options : dict, optional
            Dictionary of keyword arguments to be passed directly to tqdm.
            See https://github.com/tqdm/tqdm#parameters for options.
        """
        self.imaging_extractor = imaging_extractor

        assert not (buffer_gb and buffer_shape), "Only one of 'buffer_gb' or 'buffer_shape' can be specified!"
        assert not (chunk_mb and chunk_shape), "Only one of 'chunk_mb' or 'chunk_shape' can be specified!"

        if chunk_mb and buffer_gb:
            assert chunk_mb * 1e6 <= buffer_gb * 1e9, "chunk_mb must be less than or equal to buffer_gb!"

        if chunk_mb is None and chunk_shape is None:
            chunk_mb = 10.0

        if chunk_shape is None:
            chunk_shape = self._get_default_chunk_shape(chunk_mb=chunk_mb)

        if buffer_gb is None and buffer_shape is None:
            buffer_gb = 1.0

        if buffer_shape is None:
            buffer_shape = self._get_scaled_buffer_shape(buffer_gb=buffer_gb, chunk_shape=chunk_shape)

        super().__init__(
            buffer_shape=buffer_shape,
            chunk_shape=chunk_shape,
            display_progress=display_progress,
            progress_bar_class=progress_bar_class,
            progress_bar_options=progress_bar_options,
        )

    def _get_sample_shape(self) -> tuple:
        """This translate the sample shape in roiextractors to the nwb convention by transposing the frame shape."""

        roi_extractors_frame_shape = self.imaging_extractor.get_frame_shape()
        height, width = roi_extractors_frame_shape[0], roi_extractors_frame_shape[1]
        nwb_frame_shape = (width, height)

        if self.imaging_extractor.is_volumetric:
            num_planes = self.imaging_extractor.get_num_planes()
            sample_shape = nwb_frame_shape + (num_planes,)
        else:
            sample_shape = nwb_frame_shape

        return sample_shape

    def _get_default_chunk_shape(self, chunk_mb: float) -> tuple:
        """Select the chunk_shape less than the threshold of chunk_mb while keeping the original image size."""
        assert chunk_mb > 0, f"chunk_mb ({chunk_mb}) must be greater than zero!"

        num_samples = self.imaging_extractor.get_num_samples()
        sample_shape = self._get_sample_shape()
        dtype = self.imaging_extractor.get_dtype()

        chunk_shape = get_image_series_chunk_shape(
            num_samples=num_samples,
            sample_shape=sample_shape,
            dtype=dtype,
            chunk_mb=chunk_mb,
        )

        return chunk_shape

    def _get_scaled_buffer_shape(self, buffer_gb: float, chunk_shape: tuple) -> tuple:
        """Select the buffer_shape less than the threshold of buffer_gb that is also a multiple of the chunk_shape."""
        assert buffer_gb > 0, f"buffer_gb ({buffer_gb}) must be greater than zero!"
        assert all(np.array(chunk_shape) > 0), f"Some dimensions of chunk_shape ({chunk_shape}) are less than zero!"

        sample_shape = self._get_sample_shape()
        series_shape = self._get_maxshape()
        dtype = self._get_dtype()

        buffer_shape = get_image_series_buffer_shape(
            chunk_shape=chunk_shape,
            sample_shape=sample_shape,
            series_shape=series_shape,
            dtype=dtype,
            buffer_gb=buffer_gb,
        )

        return buffer_shape

    def _get_dtype(self) -> np.dtype:
        return self.imaging_extractor.get_dtype()

    def _get_maxshape(self) -> tuple:

        num_frames = self.imaging_extractor.get_num_samples()
        sample_shape = self._get_sample_shape()

        max_shape = (num_frames,) + sample_shape
        return max_shape

    def _get_data(self, selection: tuple[slice]) -> np.ndarray:
        data = self.imaging_extractor.get_series(
            start_sample=selection[0].start,
            end_sample=selection[0].stop,
        )
        tranpose_axes = (0, 2, 1) if len(data.shape) == 3 else (0, 2, 1, 3)
        sliced_selection = (slice(0, self.buffer_shape[0]),) + selection[1:]

        return data.transpose(tranpose_axes)[sliced_selection]



from pydantic import ConfigDict, DirectoryPath, validate_call
from pynwb import NWBFile

from ._openephys_utils import _get_session_start_time, _read_settings_xml
from ....basedatainterface import BaseDataInterface
from ....utils import (
    DeepDict,
    get_json_schema_from_method_signature,
)


class OpenEphysBinaryAnalogInterface(BaseDataInterface):
    """Primary data interface class for converting analog channels from OpenEphysBinary data."""

    display_name = "OpenEphysBinary Analog Recording"
    keywords = ("OpenEphys", "analog", "ADC")
    associated_suffixes = (".dat", ".oebin", ".npy")
    info = "Interface for OpenEphysBinary analog channel recording data."

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = get_json_schema_from_method_signature(method=cls.__init__)
        source_schema["properties"]["folder_path"]["description"] = "Path to OpenEphys directory (.dat files)."
        return source_schema

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def __init__(
        self,
        folder_path: DirectoryPath,
        stream_name: str | None = None,
        block_index: int | None = None,
        verbose: bool = False,
        time_series_name: str = "TimeSeriesOpenEphysAnalog",
    ):
        """
        Read analog channel data from the OpenEphysBinary recording.

        Parameters
        ----------
        folder_path : DirectoryPath
            Path to directory containing OpenEphys binary files.
        stream_name : str, optional
            The name of the recording stream to load; only required if there is more than one stream detected.
            Call `OpenEphysRecordingInterface.get_stream_names(folder_path=...)` to see what streams are available.
        block_index : int, optional, default: None
            The index of the block to extract from the data.
        verbose : bool, default: False
            Controls verbosity.
        time_series_name : str, default: "TimeSeriesOpenEphysAnalog"
            The name of the TimeSeries object in the NWBFile and also
            the key of the associated metadata
        """
        from spikeinterface.extractors.extractor_classes import (
            OpenEphysBinaryRecordingExtractor,
        )

        self.folder_path = folder_path
        self._xml_root = _read_settings_xml(folder_path)
        self.time_series_name = time_series_name

        available_streams = OpenEphysBinaryRecordingExtractor.get_streams(folder_path=folder_path)[0]
        non_neural_streams_indicators = ["ADC", "NI-DAQ"]
        is_non_neural = lambda stream_id: any(indicator in stream_id for indicator in non_neural_streams_indicators)
        available_non_neural_streams = [stream_id for stream_id in available_streams if is_non_neural(stream_id)]
        if len(available_non_neural_streams) > 1 and stream_name is None:
            raise ValueError(
                "More than one stream is detected! "
                "Please specify which stream you wish to load with the `stream_name` argument. "
                "To see what streams are available, call "
                " `OpenEphysRecordingInterface.get_stream_names(folder_path=...)`."
            )
        if stream_name is not None and stream_name not in available_non_neural_streams:
            raise ValueError(
                f"The selected stream '{stream_name}' is not in the available adc streams are '{available_non_neural_streams}'!"
            )

        self.stream_name = stream_name or available_streams[0]
        self.block_index = block_index

        # Initialize the recording extractor
        self.recording_extractor = OpenEphysBinaryRecordingExtractor(
            folder_path=folder_path,
            stream_name=self.stream_name,
            block_index=block_index,
        )

        # Filter for only analog channels (ADC)
        channel_ids = self.recording_extractor.get_channel_ids()
        analog_prefixes = ["ADC", "AI", "AUX"]
        is_analog = lambda ch: any(prefix in str(ch) for prefix in analog_prefixes)
        self.analog_channel_ids = [ch for ch in channel_ids if is_analog(ch)]

        if not self.analog_channel_ids:
            raise ValueError(f"No analog channels (ADC) found in the selected stream '{self.stream_name}'!")

        # Select only analog channels
        self.recording_extractor = self.recording_extractor.select_channels(channel_ids=self.analog_channel_ids)

        super().__init__(
            folder_path=folder_path,
            stream_name=stream_name,
            block_index=block_index,
            verbose=verbose,
        )

    def get_metadata(self) -> DeepDict:
        metadata = super().get_metadata()

        session_start_time = _get_session_start_time(element=self._xml_root)
        if session_start_time is not None:
            metadata["NWBFile"].update(session_start_time=session_start_time)

        return metadata

    def get_channel_names(self) -> list[str]:
        """
        Get a list of analog channel names from the recording extractor.

        Returns
        -------
        list of str
            The names of all analog channels in the OpenEphys recording.
        """
        return list(self.analog_channel_ids)

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata: dict | None = None,
        stub_test: bool = False,
        iterator_type: str | None = "v2",
        iterator_opts: dict | None = None,
        always_write_timestamps: bool = False,
    ):
        """
        Add analog channel data to an NWB file.

        Parameters
        ----------
        nwbfile : NWBFile
            The NWB file to which the analog data will be added
        metadata : dict, optional
            Metadata dictionary with device information. If None, uses default metadata
        stub_test : bool, default: False
            If True, only writes a small amount of data for testing
        iterator_type : str, optional, default: "v2"
            Type of iterator to use for data streaming
        iterator_opts : dict, optional
            Additional options for the iterator
        always_write_timestamps : bool, default: False
            If True, always writes timestamps instead of using sampling rate
        """
        from ....tools.spikeinterface import (
            _stub_recording,
            add_recording_as_time_series_to_nwbfile,
        )

        if metadata is None:
            metadata = self.get_metadata()

        recording = self.recording_extractor
        if stub_test:
            recording = _stub_recording(recording=recording)

        description = (
            f"ADC data acquired with OpenEphys system. \n Channels are {self.get_channel_names()} in that order."
        )
        metadata["TimeSeries"][self.time_series_name] = dict(name=self.time_series_name, description=description)

        add_recording_as_time_series_to_nwbfile(
            recording=recording,
            nwbfile=nwbfile,
            metadata=metadata,
            iterator_type=iterator_type,
            iterator_opts=iterator_opts,
            always_write_timestamps=always_write_timestamps,
            metadata_key=self.time_series_name,
        )



from typing import Optional

from pydantic import FilePath, validate_call

from ..baseimagingextractorinterface import BaseImagingExtractorInterface
from ....utils import DeepDict


def is_file_multiplane(file_path):
    """
    Hacky check for 'multiplane' keyword in the file.
    Reads line by line to avoid memory issues with large files.
    If found, raises NotImplementedError.
    This is NOT a proper ISX API method—just a string search.
    """
    with open(file_path, "r", encoding="utf-8", errors="ignore") as f:
        for line in f:
            if "multiplane" in line:
                raise NotImplementedError(
                    f"Multiplane ISXD file detected (found 'multiplane' in file).\n"
                    f"This is a hacky check (not an official ISX API method) and may not be robust.\n"
                    f"Proper separation logic is not yet implemented in roiextractors.\n"
                    f"Loading as 2D would result in incorrect data interpretation.\n\n"
                    f"Please open an issue at:\n"
                    f"https://github.com/catalystneuro/roiextractors/issues\n\n"
                    f"Reference: https://github.com/inscopix/pyisx/issues/36"
                )


class InscopixImagingInterface(BaseImagingExtractorInterface):
    """Data Interface for Inscopix Imaging Extractor."""

    display_name = "Inscopix Imaging"
    associated_suffixes = (".isxd",)
    info = "Interface for handling Inscopix imaging data."

    @validate_call
    def __init__(
        self,
        file_path: FilePath,
        verbose: bool = False,
        **kwargs,
    ):
        """
        Parameters
        ----------
        file_path : FilePath
            Path to the .isxd Inscopix file.
        verbose : bool, optional
            If True, outputs additional information during processing.
        **kwargs : dict, optional
            Additional keyword arguments passed to the parent class.

        Raises
        ------
        NotImplementedError
            If the file contains multiplane configuration that is not yet supported.
        """
        # Check for multiplane configuration before proceeding
        is_file_multiplane(file_path)

        kwargs.setdefault("photon_series_type", "OnePhotonSeries")
        super().__init__(
            file_path=file_path,
            verbose=verbose,
            **kwargs,
        )

    def get_metadata(self) -> DeepDict:
        """
        Retrieve the metadata for the Inscopix imaging data.

        Returns
        -------
        DeepDict
            Dictionary containing metadata including device information, imaging plane details,
            photon series configuration, and Inscopix-specific acquisition parameters.
        """
        # Get metadata from parent (already configured for OnePhotonSeries)
        metadata = super().get_metadata()

        extractor = self.imaging_extractor
        extractor_metadata = extractor._get_metadata()

        # Extract individual components
        session_info = extractor_metadata.get("session", {})
        device_info = extractor_metadata.get("device", {})
        subject_info = extractor_metadata.get("subject", {})
        session_start_time = extractor_metadata.get("session_start_time")

        # Session start time
        if session_start_time:
            metadata["NWBFile"]["session_start_time"] = session_start_time

        # Session name and experimenter
        if session_info.get("session_name"):
            metadata["NWBFile"]["session_id"] = session_info["session_name"]
        if session_info.get("experimenter_name"):
            metadata["NWBFile"]["experimenter"] = [session_info["experimenter_name"]]

        # Device information
        if device_info:
            device_metadata = metadata["Ophys"]["Device"][0]

            # Update the actual device name
            if device_info.get("device_name"):
                device_metadata["name"] = device_info["device_name"]

            # Build device description
            microscope_info = []
            if device_info.get("device_serial_number"):
                microscope_info.append(f"Serial: {device_info['device_serial_number']}")
            if device_info.get("acquisition_software_version"):
                microscope_info.append(f"Software: {device_info['acquisition_software_version']}")

            if microscope_info:
                device_metadata["description"] = f"Inscopix Microscope ({', '.join(microscope_info)})"

            # Update imaging plane metadata with acquisition details
            imaging_plane_metadata = metadata["Ophys"]["ImagingPlane"][0]

            # Update imaging plane device reference to match the actual device name
            if device_info.get("device_name"):
                imaging_plane_metadata["device"] = device_info["device_name"]

            acquisition_details = []
            if device_info.get("exposure_time_ms"):
                acquisition_details.append(f"Exposure Time (ms): {device_info['exposure_time_ms']}")
            if device_info.get("microscope_gain"):
                acquisition_details.append(f"Gain: {device_info['microscope_gain']}")
            if device_info.get("microscope_focus"):
                acquisition_details.append(f"Focus: {device_info['microscope_focus']}")
            if device_info.get("efocus"):
                acquisition_details.append(f"eFocus: {device_info['efocus']}")
            if device_info.get("led_power_ex_mw_per_mm2"):
                acquisition_details.append(f"EX LED Power (mw/mm^2): {device_info['led_power_ex_mw_per_mm2']}")
            if device_info.get("led_power_og_mw_per_mm2"):
                acquisition_details.append(f"OG LED Power (mw/mm^2): {device_info['led_power_og_mw_per_mm2']}")

            if acquisition_details:
                current_description = imaging_plane_metadata.get(
                    "description", "The plane or volume being imaged by the microscope."
                )
                imaging_plane_metadata["description"] = f"{current_description} ({'; '.join(acquisition_details)})"

        # Subject information
        subject_metadata = {}
        has_any_subject_data = False

        # Subject ID
        if subject_info and subject_info.get("animal_id"):
            subject_metadata["subject_id"] = subject_info["animal_id"]
            has_any_subject_data = True

        species_value = None
        strain_value = None

        if subject_info and subject_info.get("species"):
            species_raw = subject_info["species"]
            # If it contains genotype info or matches NWB format, put it in species; otherwise strain
            # e.g., "CaMKIICre"
            if " " in species_raw and species_raw[0].isupper() and species_raw.split()[1][0].islower():
                species_value = species_raw
            else:
                strain_value = species_raw

            if species_value:
                subject_metadata["species"] = species_value
                has_any_subject_data = True
            if strain_value:
                subject_metadata["strain"] = strain_value
                has_any_subject_data = True

        # Sex mapping
        sex_mapping = {"m": "M", "male": "M", "f": "F", "female": "F", "u": "U", "unknown": "U"}
        if subject_info and subject_info.get("sex"):
            mapped_sex = sex_mapping.get(subject_info["sex"].lower(), "U")
            subject_metadata["sex"] = mapped_sex
            has_any_subject_data = True

        # Additional subject fields
        if subject_info:
            if subject_info.get("description"):
                subject_metadata["description"] = subject_info["description"]
                has_any_subject_data = True
            if subject_info.get("date_of_birth"):
                subject_metadata["date_of_birth"] = subject_info["date_of_birth"]
                has_any_subject_data = True
            if subject_info.get("weight") and subject_info["weight"] > 0:
                subject_metadata["weight"] = str(subject_info["weight"])
                has_any_subject_data = True

        # Add Subject if we have ANY subject information, filling required fields with defaults
        if has_any_subject_data:
            if "subject_id" not in subject_metadata:
                subject_metadata["subject_id"] = "Unknown"
            if "species" not in subject_metadata:
                subject_metadata["species"] = "Unknown species"
            if "sex" not in subject_metadata:
                subject_metadata["sex"] = "U"

            if "Subject" in metadata:
                metadata["Subject"].update(subject_metadata)
            else:
                metadata["Subject"] = subject_metadata

        return metadata

    def add_to_nwbfile(
        self,
        nwbfile,
        metadata: Optional[dict] = None,
        **kwargs,
    ):
        """
        Add the Inscopix data to the NWB file.

        Parameters
        ----------
        nwbfile : NWBFile
            NWB file to add the data to.
        metadata : dict, optional
            Metadata dictionary. If None, will be generated dynamically with OnePhotonSeries.
        **kwargs
            Additional keyword arguments passed to the parent add_to_nwbfile method.

        # TODO: add logic for determining whether the microscope is `nVista 2P` and change photon_series_type to TwoPhotonSeries accordingly.
        """
        kwargs["photon_series_type"] = "OnePhotonSeries"
        # TODO: add logic for determining whether the microscope is `nVista 2P` and change photon_series_type to TwoPhotonSeries accordingly.
        super().add_to_nwbfile(nwbfile=nwbfile, metadata=metadata, **kwargs)



import datetime
import json
import warnings
from pathlib import Path
from typing import Optional

import numpy as np
from dateutil.parser import parse as dateparse
from pydantic import DirectoryPath, FilePath, validate_call

from ..baseimagingextractorinterface import BaseImagingExtractorInterface
from ....utils import DeepDict


class ScanImageImagingInterface(BaseImagingExtractorInterface):
    """Interface for reading TIFF files produced via ScanImage software.

    This interface is designed to handle the structure of ScanImage TIFF files, which can contain
    multi-channel and both planar and volumetric data. It supports both single-file and multi-file
    datasets generated by ScanImage in various acquisition modes (grab, focus, loop).

    ScanImage is a software package for controlling laser scanning microscopes, particularly for
    two-photon and multi-photon imaging. The interface extracts imaging data and metadata from
    ScanImage TIFF files and converts them to NWB format.

    Key features:

    * Handles multi-channel data with channel selection
    * Supports volumetric (multi-plane) imaging data
    * Automatically detects and loads multi-file datasets based on ScanImage naming conventions
    * Extracts and provides access to ScanImage metadata
    * Efficiently retrieves frames using lazy loading
    * Handles flyback frames in volumetric data
    """

    display_name = "ScanImage Imaging"
    associated_suffixes = (".tif", ".tiff")
    info = "Interface for ScanImage TIFF files."

    ExtractorName = "ScanImageImagingExtractor"

    @validate_call
    def __init__(
        self,
        file_path: Optional[FilePath] = None,
        channel_name: Optional[str] = None,
        slice_sample: Optional[int] = None,
        plane_index: Optional[int] = None,
        file_paths: Optional[list[FilePath]] = None,
        interleave_slice_samples: Optional[bool] = None,
        plane_name: str | None = None,
        fallback_sampling_frequency: float | None = None,
        verbose: bool = False,
    ):
        """
        Parameters
        ----------
        file_path : FilePath, optional
            Path to the ScanImage TIFF file. If this is part of a multi-file series, this should be the first file.
            Either `file_path` or `file_paths` must be provided.
        channel_name : str, optional
            Name of the channel to extract (e.g., "Channel 1", "Channel 2").

            - If None and only one channel is available, that channel will be used.
            - If None and multiple channels are available, an error will be raised.
            - Use `get_available_channels(file_path)` to see available channels before creating the interface.
        slice_sample : int, optional
            Controls how to handle multiple frames per slice in volumetric data:
            ScanImage data can contain multiple frames for a single plane. Use this to
            select a specific frame from each slice. if None, this will throw an error.
            Select a slice sample or set `interleave_slice_samples` to True to interleave
            all the slice samples as separate volumes/samples. Note that this will scramble the
            acquisition order of the frames.
            This parameter has no effect when frames_per_slice = 1.
        plane_index : int, optional
            Must be between ``0`` and ``num_planes-1``. Used to extract a specific plane from volumetric data.
            When provided:

            - The resulting extractor will be planar
            - Each sample will contain only data for the specified plane
            - This parameter has no effect on planar (non-volumetric) data.
        file_paths : list[Path | str], optional
            List of file paths to use. This is an escape value that can be used
            in case the automatic file detection doesn't work correctly and can be used
            to override the automatic file detection.
            This is useful when:

            - Automatic detection doesn't work correctly
            - You need to specify a custom subset of files
            - You need to control the exact order of files
            The file paths must be provided in the temporal order of the frames in the dataset.
        interleave_slice_samples : bool, optional
            Controls whether to interleave all slice samples as separate time points when frames_per_slice > 1:

            - If True: Interleaves all slice samples as separate time points, increasing the effective
            number of samples by frames_per_slice. This treats each slice_sample as a distinct sample.
            - If False: Requires a specific slice_sample to be provided when frames_per_slice > 1.
            - This parameter has no effect when ``frames_per_slice = 1`` or when ``slice_sample`` is provided.
            - Default is True for backward compatibility (will change to False after November 2025).
        plane_name : str, optional
            Deprecated. Use plane_index instead. Will be removed in or after November 2025.
        fallback_sampling_frequency : float, optional
            Deprecated. Will be removed in or after November 2025.
        verbose : bool, default: False
            If True, will print detailed information about the interface initialization process.
        """
        file_paths = [Path(file_path)] if file_path else file_paths
        header_version = self.get_scanimage_version(file_path=file_paths[0])
        if header_version not in [3, 4, 5]:
            raise ValueError(
                f"Unsupported ScanImage version {header_version}. Supported versions are 3, 4, and 5."
                f"Most likely this is a legacy version, use ScanImageLegacyImagingInterface instead."
            )

        # Backward compatibility flag - will be set to False after November 2025
        if interleave_slice_samples is None:
            interleave_slice_samples = True
            warnings.warn(
                "interleave_slice_samples currently set to True for backward compatibility. \n"
                "This will be set to False by default in or after November 2025."
            )

        if plane_name is not None:

            warnings.warn(
                "The `plane_name` argument is deprecated and will be removed in or after November 2025. Use `plane_index` instead."
            )
            plane_index = int(plane_name)

        if fallback_sampling_frequency is not None:
            warnings.warn(
                "The `fallback_sampling_frequency` argument is deprecated and will be removed in or after November 2025"
            )

        self.channel_name = channel_name
        self.plane_index = plane_index
        super().__init__(
            file_path=file_path,
            channel_name=channel_name,
            file_paths=file_paths,
            plane_index=plane_index,
            slice_sample=slice_sample,
            interleave_slice_samples=interleave_slice_samples,
            verbose=verbose,
        )

        # Make sure the timestamps are available, the extractor caches them
        times = self.imaging_extractor.get_times()
        self.imaging_extractor.set_times(times=times)

    def get_metadata(self) -> DeepDict:
        """
        Get metadata for the ScanImage imaging data.

        Returns
        -------
        DeepDict
            The metadata dictionary containing imaging metadata from the ScanImage files.
            This includes:
            - Session start time extracted from the ScanImage file
            - Device information for the microscope
            - Optical channel configuration
            - Imaging plane details including grid spacing and origin coordinates if available
            - Photon series metadata with scan line rate and other acquisition parameters
        """
        metadata = super().get_metadata()

        session_start_time = self._get_session_start_time()
        if session_start_time:
            metadata["NWBFile"]["session_start_time"] = session_start_time

        # Extract ScanImage-specific metadata
        if hasattr(self.imaging_extractor, "_general_metadata"):
            # Add general metadata to a custom field
            scanimage_metadata = self.imaging_extractor._general_metadata

            # Update device information
            device_name = "Microscope"
            metadata["Ophys"]["Device"][0].update(name=device_name, description=f"Microscope controlled by ScanImage")
            channel_name_string = self.channel_name.replace(" ", "").capitalize()

            optical_channel_name = f"OpticalChannel{channel_name_string}"
            optical_channel_metadata = {
                "name": optical_channel_name,
                "description": "Optical channel from ScanImage acquisition",
                "emission_lambda": np.nan,
            }

            # Update imaging plane metadata
            imaging_plane_metadata = metadata["Ophys"]["ImagingPlane"][0]
            plane_index_string = f"Plane{self.plane_index}" if self.plane_index is not None else ""
            imaging_plane_name = f"ImagingPlane{channel_name_string}{plane_index_string}"
            imaging_plane_metadata.update(
                name=imaging_plane_name,
                device=device_name,
                imaging_rate=self.imaging_extractor.get_sampling_frequency(),
                description="Imaging plane from ScanImage acquisition",
                optical_channel=[optical_channel_metadata],
            )

            # Update photon series metadata
            photon_series_key = self.photon_series_type  # "TwoPhotonSeries" or "OnePhotonSeries"
            photon_series_metadata = metadata["Ophys"][photon_series_key][0]

            photon_series_name = f"{photon_series_key}{channel_name_string}{plane_index_string}"
            photon_series_metadata["imaging_plane"] = imaging_plane_name
            photon_series_metadata["name"] = photon_series_name
            photon_series_metadata["description"] = f"Imaging data acquired using ScanImage for {self.channel_name}"

            # Add additional metadata if available
            if "FrameData" in scanimage_metadata:
                frame_data = scanimage_metadata["FrameData"]

                # Calculate scan line rate from line period if available
                if "SI.hRoiManager.linePeriod" in frame_data:
                    scan_line_rate = 1 / float(frame_data["SI.hRoiManager.linePeriod"])
                    photon_series_metadata.update(scan_line_rate=scan_line_rate)
                elif "SI.hScan2D.scannerFrequency" in frame_data:
                    photon_series_metadata.update(scan_line_rate=frame_data["SI.hScan2D.scannerFrequency"])

                # Add version information to device description if available
                if "SI.VERSION_MAJOR" in frame_data:
                    version = f"{frame_data.get('SI.VERSION_MAJOR', '')}.{frame_data.get('SI.VERSION_MINOR', '')}.{frame_data.get('SI.VERSION_UPDATE', '')}"
                    metadata["Ophys"]["Device"][0][
                        "description"
                    ] = f"Microscope and acquisition data with ScanImage (version {version})"

            # Extract ROI metadata if available
            if "RoiGroups" in scanimage_metadata:
                roi_metadata = scanimage_metadata["RoiGroups"]

                # Extract grid spacing and origin coordinates from scanfields
                grid_spacing = None
                grid_spacing_unit = "n.a"
                origin_coords = None
                origin_coords_unit = "n.a"

                if "imagingRoiGroup" in roi_metadata and "rois" in roi_metadata["imagingRoiGroup"]:
                    rois = roi_metadata["imagingRoiGroup"]["rois"]
                    if isinstance(rois, dict) and "scanfields" in rois:
                        scanfields = rois["scanfields"]
                        if "sizeXY" in scanfields and "pixelResolutionXY" in scanfields:
                            fov_size_in_um = np.array(scanfields["sizeXY"])
                            frame_dimension = np.array(scanfields["pixelResolutionXY"])
                            grid_spacing = fov_size_in_um / frame_dimension
                            grid_spacing_unit = "micrometers"

                        if "centerXY" in scanfields:
                            origin_coords = scanfields["centerXY"]
                            origin_coords_unit = "micrometers"

                # Update imaging plane metadata with grid spacing and origin coordinates
                if grid_spacing is not None:
                    imaging_plane_metadata.update(
                        grid_spacing=grid_spacing.tolist(), grid_spacing_unit=grid_spacing_unit
                    )

                if origin_coords is not None:
                    imaging_plane_metadata.update(origin_coords=origin_coords, origin_coords_unit=origin_coords_unit)

        return metadata

    def _get_session_start_time(self) -> datetime.datetime | None:
        """
        Extract and parse the 'epoch' metadata from a ScanImage TIFF file as the session start time.

        This method opens the first frame of the TIFF file, extracts the 'epoch' field from the
        ImageDescription tag, and parses it into a datetime object.

        Returns
        -------
        datetime
            Parsed datetime from the 'epoch' metadata.

        Raises
        ------
        ValueError
            If 'epoch' metadata is not found in the file.
        """

        from tifffile import TiffReader

        tiff_file_path = self.imaging_extractor.file_path
        with TiffReader(tiff_file_path) as tif:
            image_description = tif.pages[0].tags["ImageDescription"].value

        import re

        match = re.search(r"epoch\s*=\s*\[([^\]]+)\]", image_description)
        if not match:
            raise ValueError(f"'epoch' field not found in {tiff_file_path}")

        epoch_values = match.group(1).split()
        import warnings

        if len(epoch_values) != 6:
            warnings.warn(
                f"Expected 6 values in 'epoch' field, found {len(epoch_values)}: \n" f"Epoch field {epoch_values}."
            )
            return None

        year, month, day, hour, minute, seconds = map(float, epoch_values)
        second_int = int(seconds)
        microsecond = int((seconds - second_int) * 1e6)

        return datetime.datetime(int(year), int(month), int(day), int(hour), int(minute), second_int, microsecond)

    @staticmethod
    def get_scanimage_version(file_path: Path | str) -> int:
        """
        Extract the ScanImage version from a BigTIFF file without validation.

        This method reads the binary header of the TIFF file to determine the ScanImage version
        that produced it. It supports ScanImage versions 3, 4, and 5.

        Parameters
        ----------
        file_path : Path | str
            Path to the ScanImage TIFF file

        Returns
        -------
        int
            ScanImage version number (3, 4, or 5)
        """
        with open(file_path, "rb") as f:
            # Skip the TIFF header (16 bytes) and the Magic Number (4 bytes)
            f.seek(20)

            # Read ScanImage version (4 bytes)
            version_bytes = f.read(4)
            scanimage_version = int.from_bytes(version_bytes, byteorder="little")

            return scanimage_version

    @staticmethod
    def get_available_channels(file_path: Path | str) -> list[str]:
        """
        Get the channel names available in a ScanImage TIFF file.

        This static method extracts the channel names from a ScanImage TIFF file
        without needing to create an interface instance. This is useful for
        determining which channels are available before creating an interface.

        Parameters
        ----------
        file_path : Path | str
            Path to the ScanImage TIFF file.

        Returns
        -------
        list[str]
            List of channel names available in the file (e.g., ["Channel 1", "Channel 2"]).
        """
        from roiextractors import ScanImageImagingExtractor

        return ScanImageImagingExtractor.get_available_channels(file_path=file_path)

    @staticmethod
    def get_available_planes(file_path: Path | str) -> list[str]:
        """
        Get the available plane names from a ScanImage TIFF file.

        This static method determines the number of planes (Z-slices) in a volumetric
        ScanImage dataset without needing to create an interface instance. This is useful
        for determining which planes are available before creating an interface.

        Parameters
        ----------
        file_path : Path | str
            Path to the ScanImage TIFF file.

        Returns
        -------
        list[str]
            List of plane names available in the file. For volumetric data, this will be
            a list of strings representing plane indices (e.g., ["0", "1", "2"]).
        """
        from roiextractors import ScanImageImagingExtractor

        return ScanImageImagingExtractor.get_available_planes(file_path=file_path)


class ScanImageLegacyImagingInterface(BaseImagingExtractorInterface):
    """Interface for reading TIFF files produced via ScanImage v3.8."""

    display_name = "ScanImage Imaging"
    associated_suffixes = (".tif",)
    info = "Interface for ScanImage v3.8 TIFF files."

    ExtractorName = "ScanImageLegacyImagingExtractor"

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = super().get_source_schema()
        source_schema["properties"]["file_path"]["description"] = "Path to Tiff file."
        return source_schema

    def _source_data_to_extractor_kwargs(self, source_data: dict) -> dict:
        extractor_kwargs = source_data.copy()
        extractor_kwargs.pop("fallback_sampling_frequency", None)
        extractor_kwargs["sampling_frequency"] = self.sampling_frequency

        return extractor_kwargs

    @validate_call
    def __init__(
        self,
        file_path: FilePath,
        fallback_sampling_frequency: float | None = None,
        verbose: bool = False,
    ):
        """
        DataInterface for reading Tiff files that are generated by ScanImage v3.8. This interface extracts the metadata
        from the exif of the tiff file.

        Parameters
        ----------
        file_path: FilePath
            Path to tiff file.
        fallback_sampling_frequency: float, optional
            The sampling frequency can usually be extracted from the scanimage metadata in
            exif:ImageDescription:state.acq.frameRate. If not, use this.
        """
        from roiextractors.extractors.tiffimagingextractors.scanimagetiff_utils import (
            extract_extra_metadata,
        )

        self.image_metadata = extract_extra_metadata(file_path=file_path)

        if "state.acq.frameRate" in self.image_metadata:
            sampling_frequency = float(self.image_metadata["state.acq.frameRate"])
        elif "SI.hRoiManager.scanFrameRate" in self.image_metadata:
            sampling_frequency = float(self.image_metadata["SI.hRoiManager.scanFrameRate"])
        else:
            assert_msg = (
                "sampling frequency not found in image metadata, "
                "input the frequency using the argument `fallback_sampling_frequency`"
            )
            assert fallback_sampling_frequency is not None, assert_msg
            sampling_frequency = fallback_sampling_frequency

        self.sampling_frequency = sampling_frequency
        super().__init__(file_path=file_path, fallback_sampling_frequency=fallback_sampling_frequency, verbose=verbose)

    def get_metadata(self) -> DeepDict:
        """
        Get metadata for the ScanImage imaging data.

        Returns
        -------
        dict
            Dictionary containing metadata including session start time and device information
            specific to the ScanImage system.
        """
        device_number = 0  # Imaging plane metadata is a list with metadata for each plane

        metadata = super().get_metadata()

        if "state.internal.triggerTimeString" in self.image_metadata:
            extracted_session_start_time = dateparse(self.image_metadata["state.internal.triggerTimeString"])
            metadata["NWBFile"].update(session_start_time=extracted_session_start_time)

        # Extract many scan image properties and attach them as dic in the description
        ophys_metadata = metadata["Ophys"]
        two_photon_series_metadata = ophys_metadata["TwoPhotonSeries"][device_number]
        if self.image_metadata is not None:
            extracted_description = json.dumps(self.image_metadata)
            two_photon_series_metadata.update(description=extracted_description)

        return metadata


class ScanImageMultiFileImagingInterface(BaseImagingExtractorInterface):
    """
    Interface for reading multi-file (buffered) TIFF files produced via ScanImage.

    It extracts metadata from the first TIFF file in the provided folder and determines the number of available planes.
    If there is more than one plane and no specific plane is provided, it creates an instance of ScanImageMultiPlaneMultiFileImagingInterface.
    If there is only one plane or a specific plane is provided, it creates an instance of ScanImageSinglePlaneMultiFileImagingInterface.
    """

    display_name = "ScanImage Multi-File Imaging"
    associated_suffixes = (".tif",)
    info = "Interface for ScanImage multi-file (buffered) TIFF files."

    ExtractorName = "ScanImageTiffSinglePlaneMultiFileImagingExtractor"

    @classmethod
    def get_source_schema(cls) -> dict:
        """
        Get the source schema for the ScanImage multi-file imaging interface.

        Returns
        -------
        dict
            The schema dictionary containing input parameters and descriptions
            for initializing the ScanImage multi-file interface.
        """
        source_schema = super().get_source_schema()
        source_schema["properties"]["folder_path"]["description"] = "Path to the folder containing the TIFF files."
        return source_schema

    @validate_call
    def __new__(
        cls,
        folder_path: DirectoryPath,
        file_pattern: str,
        channel_name: str | None = None,
        plane_name: str | None = None,
        extract_all_metadata: bool = False,
        verbose: bool = False,
    ):
        warnings.warn(
            "ScanImageMultiFileImagingInterface is deprecated and will be removed in or after October 2025. "
            "Use ScanImageImagingInterface instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        from natsort import natsorted
        from roiextractors.extractors.tiffimagingextractors.scanimagetiff_utils import (
            extract_extra_metadata,
            parse_metadata,
        )

        file_paths = natsorted(Path(folder_path).glob(file_pattern))
        first_file_path = file_paths[0]

        image_metadata = extract_extra_metadata(file_path=first_file_path)

        version = get_scanimage_major_version(scanimage_metadata=image_metadata)
        if version == "3.8":
            raise ValueError("ScanImage version 3.8 is not supported.")

        parsed_metadata = parse_metadata(metadata=image_metadata)
        available_planes = [f"{i}" for i in range(parsed_metadata["num_planes"])]
        if len(available_planes) > 1 and plane_name is None:
            return ScanImageMultiPlaneMultiFileImagingInterface(
                folder_path=folder_path,
                file_pattern=file_pattern,
                channel_name=channel_name,
                image_metadata=image_metadata,
                parsed_metadata=parsed_metadata,
                extract_all_metadata=extract_all_metadata,
                verbose=verbose,
            )

        return ScanImageSinglePlaneMultiFileImagingInterface(
            folder_path=folder_path,
            file_pattern=file_pattern,
            channel_name=channel_name,
            plane_name=plane_name,
            image_metadata=image_metadata,
            parsed_metadata=parsed_metadata,
            extract_all_metadata=extract_all_metadata,
            verbose=verbose,
        )


class ScanImageMultiPlaneImagingInterface(BaseImagingExtractorInterface):
    """Interface for reading multi plane (volumetric) TIFF files produced via ScanImage."""

    display_name = "ScanImage Volumetric Imaging"
    associated_suffixes = (".tif",)
    info = "Interface for ScanImage multi plane (volumetric) TIFF files."

    ExtractorName = "ScanImageTiffMultiPlaneImagingExtractor"

    def _source_data_to_extractor_kwargs(self, source_data: dict) -> dict:
        extractor_kwargs = source_data.copy()
        extractor_kwargs.pop("image_metadata")
        extractor_kwargs["metadata"] = self.image_metadata

        return extractor_kwargs

    @validate_call
    def __init__(
        self,
        file_path: FilePath,
        channel_name: str | None = None,
        image_metadata: dict | None = None,
        parsed_metadata: dict | None = None,
        verbose: bool = False,
    ):
        warnings.warn(
            "ScanImageMultiPlaneImagingInterface is deprecated and will be removed in or after October 2025. "
            "Use ScanImageImagingInterface instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        """
        DataInterface for reading multi-file (buffered) TIFF files that are generated by ScanImage.

        Parameters
        ----------
        file_path : FilePath
            Path to the TIFF file.
        channel_name : str
            Name of the channel for this extractor.
        """
        from roiextractors.extractors.tiffimagingextractors.scanimagetiff_utils import (
            extract_extra_metadata,
            parse_metadata,
        )

        image_metadata = image_metadata or extract_extra_metadata(file_path=file_path)
        self.image_metadata = image_metadata
        parsed_metadata = parsed_metadata or parse_metadata(metadata=self.image_metadata)

        if parsed_metadata["num_planes"] == 1:
            raise ValueError(
                "Only one plane detected. For single plane imaging data use ScanImageSinglePlaneImagingInterface instead."
            )

        available_channels = parsed_metadata["channel_names"]
        if channel_name is None:
            if len(available_channels) > 1:
                raise ValueError(
                    "More than one channel is detected! \n "
                    "Please specify which channel you wish to load with the `channel_name` argument \n "
                    f"Available channels are: {available_channels}"
                )
            channel_name = available_channels[0]
        assert (
            channel_name in available_channels
        ), f"Channel '{channel_name}' not found! \n Available channels are: {available_channels}"

        two_photon_series_name_suffix = None
        if len(available_channels) > 1:
            two_photon_series_name_suffix = f"{channel_name.replace(' ', '')}"
        self.two_photon_series_name_suffix = two_photon_series_name_suffix

        self.metadata = image_metadata
        self.parsed_metadata = parsed_metadata
        super().__init__(
            file_path=file_path,
            channel_name=channel_name,
            image_metadata=image_metadata,
            parsed_metadata=parsed_metadata,
            verbose=verbose,
        )

    def get_metadata(self) -> DeepDict:
        """
        Get metadata for the ScanImage imaging data.

        Returns
        -------
        dict
            Dictionary containing metadata including session start time and device information
            specific to the ScanImage system.
        """
        metadata = super().get_metadata()

        extracted_session_start_time = datetime.datetime.strptime(
            self.image_metadata["epoch"], "[%Y %m %d %H %M %S.%f]"
        )
        metadata["NWBFile"].update(session_start_time=extracted_session_start_time)

        ophys_metadata = metadata["Ophys"]
        two_photon_series_metadata = ophys_metadata["TwoPhotonSeries"][0]

        if self.image_metadata is not None:
            extracted_description = json.dumps(self.image_metadata)
            two_photon_series_metadata.update(description=extracted_description)

        if self.two_photon_series_name_suffix is None:
            return metadata

        imaging_plane_metadata = ophys_metadata["ImagingPlane"][0]
        channel_name = self.source_data["channel_name"]
        optical_channel_metadata = [
            channel for channel in imaging_plane_metadata["optical_channel"] if channel["name"] == channel_name
        ]
        imaging_plane_name = f"ImagingPlane{self.two_photon_series_name_suffix}"
        imaging_plane_metadata.update(
            name=imaging_plane_name,
            optical_channel=optical_channel_metadata,
        )
        two_photon_series_metadata.update(
            name=f"TwoPhotonSeries{self.two_photon_series_name_suffix}",
            imaging_plane=imaging_plane_name,
        )

        return metadata


class ScanImageMultiPlaneMultiFileImagingInterface(BaseImagingExtractorInterface):
    """Interface for reading volumetric multi-file (buffered) TIFF files produced via ScanImage."""

    display_name = "ScanImage Volumetric Multi-File Imaging"
    associated_suffixes = (".tif",)
    info = "Interface for ScanImage multi-file (buffered) volumetric TIFF files."

    ExtractorName = "ScanImageTiffMultiPlaneMultiFileImagingExtractor"

    @validate_call
    def __init__(
        self,
        folder_path: DirectoryPath,
        file_pattern: str,
        channel_name: str | None = None,
        extract_all_metadata: bool = False,
        image_metadata: dict | None = None,
        parsed_metadata: dict | None = None,
        verbose: bool = False,
    ):
        warnings.warn(
            "ScanImageMultiPlaneMultiFileImagingInterface is deprecated and will be removed in or after October 2025. "
            "Use ScanImageImagingInterface instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        """
        DataInterface for reading multi-file (buffered) TIFF files that are generated by ScanImage.

        Parameters
        ----------
        folder_path : DirectoryPath
            Path to the folder containing the TIFF files.
        file_pattern : str
            Pattern for the TIFF files to read -- see pathlib.Path.glob for details.
        channel_name : str
            The name of the channel to load, to determine what channels are available use ScanImageTiffSinglePlaneImagingExtractor.get_available_channels(file_path=...).
        extract_all_metadata : bool
            If True, extract metadata from every file in the folder. If False, only extract metadata from the first
            file in the folder. The default is False.
        """
        from natsort import natsorted
        from roiextractors.extractors.tiffimagingextractors.scanimagetiff_utils import (
            extract_extra_metadata,
            parse_metadata,
        )

        file_paths = natsorted(Path(folder_path).glob(file_pattern))
        first_file_path = file_paths[0]

        image_metadata = image_metadata or extract_extra_metadata(file_path=first_file_path)
        self.image_metadata = image_metadata

        version = get_scanimage_major_version(scanimage_metadata=image_metadata)
        if version == "3.8":
            raise ValueError(
                "ScanImage version 3.8 is not supported. \n " "Please use ScanImageImagingInterface instead."
            )

        parsed_metadata = parsed_metadata or parse_metadata(metadata=image_metadata)
        if parsed_metadata["num_planes"] == 1:
            raise ValueError(
                "Only one plane detected. For single plane imaging data use ScanImageSinglePlaneMultiFileImagingInterface instead."
            )

        available_channels = parsed_metadata["channel_names"]
        if channel_name is None:
            if len(available_channels) > 1:
                raise ValueError(
                    "More than one channel is detected! \n "
                    "Please specify which channel you wish to load with the `channel_name` argument \n "
                    f"Available channels are: {available_channels}"
                )
            channel_name = available_channels[0]
        assert (
            channel_name in available_channels
        ), f"Channel '{channel_name}' not found! \n Available channels are: {available_channels}"

        two_photon_series_name_suffix = None
        if len(available_channels) > 1:
            two_photon_series_name_suffix = f"{channel_name.replace(' ', '')}"
        self.two_photon_series_name_suffix = two_photon_series_name_suffix

        super().__init__(
            folder_path=folder_path,
            file_pattern=file_pattern,
            channel_name=channel_name,
            extract_all_metadata=extract_all_metadata,
            verbose=verbose,
        )

    def get_metadata(self) -> DeepDict:
        """
        Get metadata for the ScanImage imaging data.

        Returns
        -------
        dict
            Dictionary containing metadata including session start time, device information,
            and imaging plane configuration specific to the ScanImage system.
        """
        metadata = super().get_metadata()

        extracted_session_start_time = datetime.datetime.strptime(
            self.image_metadata["epoch"], "[%Y %m %d %H %M %S.%f]"
        )
        metadata["NWBFile"].update(session_start_time=extracted_session_start_time)

        ophys_metadata = metadata["Ophys"]
        two_photon_series_metadata = ophys_metadata["TwoPhotonSeries"][0]

        if self.image_metadata is not None:
            extracted_description = json.dumps(self.image_metadata)
            two_photon_series_metadata.update(description=extracted_description)

        if self.two_photon_series_name_suffix is None:
            return metadata

        imaging_plane_metadata = ophys_metadata["ImagingPlane"][0]
        channel_name = self.source_data["channel_name"]
        optical_channel_metadata = [
            channel for channel in imaging_plane_metadata["optical_channel"] if channel["name"] == channel_name
        ]
        imaging_plane_name = f"ImagingPlane{self.two_photon_series_name_suffix}"
        imaging_plane_metadata.update(
            name=imaging_plane_name,
            optical_channel=optical_channel_metadata,
        )
        two_photon_series_metadata.update(
            name=f"TwoPhotonSeries{self.two_photon_series_name_suffix}",
            imaging_plane=imaging_plane_name,
        )

        return metadata


class ScanImageSinglePlaneImagingInterface(BaseImagingExtractorInterface):
    """Interface for reading TIFF files produced via ScanImage."""

    display_name = "ScanImage Single Plane Imaging"
    associated_suffixes = (".tif",)
    info = "Interface for ScanImage TIFF files."

    ExtractorName = "ScanImageTiffSinglePlaneImagingExtractor"

    def _source_data_to_extractor_kwargs(self, source_data: dict) -> dict:
        extractor_kwargs = source_data.copy()
        extractor_kwargs.pop("image_metadata")
        extractor_kwargs["metadata"] = self.image_metadata

        return extractor_kwargs

    @validate_call
    def __init__(
        self,
        file_path: FilePath,
        channel_name: str | None = None,
        plane_name: str | None = None,
        image_metadata: dict | None = None,
        parsed_metadata: dict | None = None,
        verbose: bool = False,
    ):
        warnings.warn(
            "ScanImageSinglePlaneImagingInterface is deprecated and will be removed in or after October 2025. "
            "Use ScanImageImagingInterface instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        """
        DataInterface for reading multi-file (buffered) TIFF files that are generated by ScanImage.

        Parameters
        ----------
        file_path : FilePath
            Path to the TIFF file.
        channel_name : str
            The name of the channel to load, to determine what channels are available use ScanImageTiffSinglePlaneImagingExtractor.get_available_channels(file_path=...).
        plane_name : str
            The name of the plane to load, to determine what planes are available use ScanImageTiffSinglePlaneImagingExtractor.get_available_planes(file_path=...).
        """
        from roiextractors.extractors.tiffimagingextractors.scanimagetiff_utils import (
            extract_extra_metadata,
            parse_metadata,
        )

        image_metadata = image_metadata or extract_extra_metadata(file_path=file_path)
        self.image_metadata = image_metadata

        version = get_scanimage_major_version(scanimage_metadata=image_metadata)
        if version == "3.8":
            raise ValueError(
                "ScanImage version 3.8 is not supported. \n " "Please use ScanImageImagingInterface instead."
            )

        parsed_metadata = parsed_metadata or parse_metadata(metadata=image_metadata)
        available_channels = parsed_metadata["channel_names"]
        if channel_name is None:
            if len(available_channels) > 1:
                raise ValueError(
                    "More than one channel is detected! \n "
                    "Please specify which channel you wish to load with the `channel_name` argument \n "
                    f"Available channels are: {available_channels}"
                )
            channel_name = available_channels[0]
        assert (
            channel_name in available_channels
        ), f"Channel '{channel_name}' not found! \n Available channels are: {available_channels}"

        available_planes = [f"{i}" for i in range(parsed_metadata["num_planes"])]
        if plane_name is None:
            if len(available_planes) > 1:
                raise ValueError(
                    "More than one plane is detected! \n "
                    "Please specify which plane you wish to load with the `plane_name` argument \n "
                    f"Available planes are: {available_planes}"
                )
            plane_name = available_planes[0]
        assert (
            plane_name in available_planes
        ), f"Plane '{plane_name}' not found! \n Available planes are: {available_planes}"

        two_photon_series_name_suffix = None
        if len(available_channels) > 1:
            two_photon_series_name_suffix = f"{channel_name.replace(' ', '')}"
        if len(available_planes) > 1:
            two_photon_series_name_suffix = f"{two_photon_series_name_suffix}Plane{plane_name}"
        self.two_photon_series_name_suffix = two_photon_series_name_suffix

        self.metadata = image_metadata
        self.parsed_metadata = parsed_metadata
        super().__init__(
            file_path=file_path,
            channel_name=channel_name,
            plane_name=plane_name,
            image_metadata=image_metadata,
            parsed_metadata=parsed_metadata,
            verbose=verbose,
        )

    def get_metadata(self) -> DeepDict:
        """
        Get metadata for the ScanImage imaging data.

        Returns
        -------
        dict
            Dictionary containing metadata including session start time, device information,
            and imaging plane configuration specific to the ScanImage system.
        """
        metadata = super().get_metadata()

        extracted_session_start_time = datetime.datetime.strptime(
            self.image_metadata["epoch"], "[%Y %m %d %H %M %S.%f]"
        )
        metadata["NWBFile"].update(session_start_time=extracted_session_start_time)

        ophys_metadata = metadata["Ophys"]
        two_photon_series_metadata = ophys_metadata["TwoPhotonSeries"][0]

        if self.image_metadata is not None:
            extracted_description = json.dumps(self.image_metadata)
            two_photon_series_metadata.update(description=extracted_description)

        if self.two_photon_series_name_suffix is None:
            return metadata

        imaging_plane_metadata = ophys_metadata["ImagingPlane"][0]
        channel_name = self.source_data["channel_name"]
        optical_channel_metadata = [
            channel for channel in imaging_plane_metadata["optical_channel"] if channel["name"] == channel_name
        ]
        imaging_plane_name = f"ImagingPlane{self.two_photon_series_name_suffix}"
        imaging_plane_metadata.update(
            name=imaging_plane_name,
            optical_channel=optical_channel_metadata,
        )
        two_photon_series_metadata.update(
            name=f"TwoPhotonSeries{self.two_photon_series_name_suffix}",
            imaging_plane=imaging_plane_name,
        )

        return metadata


class ScanImageSinglePlaneMultiFileImagingInterface(BaseImagingExtractorInterface):
    """Interface for reading multi-file (buffered) TIFF files produced via ScanImage."""

    display_name = "ScanImage Single Plane Multi-File Imaging"
    associated_suffixes = (".tif",)
    info = "Interface for ScanImage multi-file (buffered) TIFF files."

    ExtractorName = "ScanImageTiffSinglePlaneMultiFileImagingExtractor"

    @validate_call
    def __init__(
        self,
        folder_path: DirectoryPath,
        file_pattern: str,
        channel_name: str | None = None,
        plane_name: str | None = None,
        image_metadata: dict | None = None,
        parsed_metadata: dict | None = None,
        extract_all_metadata: bool = False,
        verbose: bool = False,
    ):
        warnings.warn(
            "ScanImageSinglePlaneMultiFileImagingInterface is deprecated and will be removed in or after October 2025. "
            "Use ScanImageImagingInterface instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        """
        DataInterface for reading multi-file (buffered) TIFF files that are generated by ScanImage.

        Parameters
        ----------
        folder_path : DirectoryPath
            Path to the folder containing the TIFF files.
        file_pattern : str
            Pattern for the TIFF files to read -- see pathlib.Path.glob for details.
        channel_name : str
            The name of the channel to load, to determine what channels are available use ScanImageTiffSinglePlaneImagingExtractor.get_available_channels(file_path=...).
        plane_name : str
            The name of the plane to load, to determine what planes are available use ScanImageTiffSinglePlaneImagingExtractor.get_available_planes(file_path=...).
        extract_all_metadata : bool
            If True, extract metadata from every file in the folder. If False, only extract metadata from the first
            file in the folder. The default is False.
        """
        from natsort import natsorted
        from roiextractors.extractors.tiffimagingextractors.scanimagetiff_utils import (
            extract_extra_metadata,
            parse_metadata,
        )

        file_paths = natsorted(Path(folder_path).glob(file_pattern))
        first_file_path = file_paths[0]

        image_metadata = image_metadata or extract_extra_metadata(file_path=first_file_path)
        self.image_metadata = image_metadata

        version = get_scanimage_major_version(scanimage_metadata=image_metadata)
        if version == "3.8":
            raise ValueError(
                "ScanImage version 3.8 is not supported. \n " "Please use ScanImageImagingInterface instead."
            )

        parsed_metadata = parsed_metadata or parse_metadata(metadata=image_metadata)
        available_channels = parsed_metadata["channel_names"]
        if channel_name is None:
            if len(available_channels) > 1:
                raise ValueError(
                    "More than one channel is detected! \n "
                    "Please specify which channel you wish to load with the `channel_name` argument \n "
                    f"Available channels are: {available_channels}"
                )
            channel_name = available_channels[0]
        assert (
            channel_name in available_channels
        ), f"Channel '{channel_name}' not found! \n Available channels are: {available_channels}"

        available_planes = [f"{i}" for i in range(parsed_metadata["num_planes"])]
        if plane_name is None:
            if len(available_planes) > 1:
                raise ValueError(
                    "More than one plane is detected! \n "
                    "Please specify which plane you wish to load with the `plane_name` argument \n "
                    f"Available planes are: {available_planes}"
                )
            plane_name = available_planes[0]
        assert (
            plane_name in available_planes
        ), f"Plane '{plane_name}' not found! \n Available planes are: {available_planes}"

        two_photon_series_name_suffix = None
        if len(available_channels) > 1:
            two_photon_series_name_suffix = f"{channel_name.replace(' ', '')}"
        if len(available_planes) > 1:
            two_photon_series_name_suffix = f"{two_photon_series_name_suffix}Plane{plane_name}"
        self.two_photon_series_name_suffix = two_photon_series_name_suffix

        super().__init__(
            folder_path=folder_path,
            file_pattern=file_pattern,
            channel_name=channel_name,
            plane_name=plane_name,
            extract_all_metadata=extract_all_metadata,
            verbose=verbose,
        )

    def get_metadata(self) -> DeepDict:
        """
        Get metadata for the ScanImage imaging data.

        Returns
        -------
        dict
            Dictionary containing metadata including session start time, device information,
            and imaging plane configuration specific to the ScanImage system.
        """
        metadata = super().get_metadata()

        extracted_session_start_time = datetime.datetime.strptime(
            self.image_metadata["epoch"], "[%Y %m %d %H %M %S.%f]"
        )
        metadata["NWBFile"].update(session_start_time=extracted_session_start_time)

        ophys_metadata = metadata["Ophys"]
        two_photon_series_metadata = ophys_metadata["TwoPhotonSeries"][0]

        if self.image_metadata is not None:
            extracted_description = json.dumps(self.image_metadata)
            two_photon_series_metadata.update(description=extracted_description)

        if self.two_photon_series_name_suffix is None:
            return metadata

        imaging_plane_metadata = ophys_metadata["ImagingPlane"][0]
        channel_name = self.source_data["channel_name"]
        optical_channel_metadata = [
            channel for channel in imaging_plane_metadata["optical_channel"] if channel["name"] == channel_name
        ]
        imaging_plane_name = f"ImagingPlane{self.two_photon_series_name_suffix}"
        imaging_plane_metadata.update(
            name=imaging_plane_name,
            optical_channel=optical_channel_metadata,
        )
        two_photon_series_metadata.update(
            name=f"TwoPhotonSeries{self.two_photon_series_name_suffix}",
            imaging_plane=imaging_plane_name,
        )

        return metadata


def get_scanimage_major_version(scanimage_metadata: dict) -> str:
    """
    Determine the version of ScanImage that produced the TIFF file.

    Parameters
    ----------
    scanimage_metadata : dict
        Dictionary of metadata extracted from a TIFF file produced via ScanImage.

    Returns
    -------
    version: str
        The version of ScanImage that produced the TIFF file.

    Raises
    ------
    ValueError
        If the ScanImage version could not be determined from metadata.
    """
    if "SI.VERSION_MAJOR" in scanimage_metadata:
        return scanimage_metadata["SI.VERSION_MAJOR"]
    elif "state.software.version" in scanimage_metadata:
        return scanimage_metadata["state.software.version"]

    raise ValueError("ScanImage version could not be determined from metadata.")



import warnings
from typing import Literal

import numpy as np
from pynwb import NWBFile
from pynwb.device import Device
from pynwb.ophys import Fluorescence, ImageSegmentation, ImagingPlane, TwoPhotonSeries

from ...baseextractorinterface import BaseExtractorInterface
from ...utils import (
    DeepDict,
    fill_defaults,
    get_base_schema,
    get_schema_from_hdmf_class,
)


class BaseSegmentationExtractorInterface(BaseExtractorInterface):
    """Parent class for all SegmentationExtractorInterfaces."""

    keywords = ("segmentation", "roi", "cells")

    ExtractorModuleName = "roiextractors"

    def __init__(self, verbose: bool = False, **source_data):
        super().__init__(**source_data)
        self.verbose = verbose
        self.segmentation_extractor = self._extractor_instance

    def get_metadata_schema(self) -> dict:
        """
        Generate the metadata schema for Ophys data, updating required fields and properties.

        This method builds upon the base schema and customizes it for Ophys-specific metadata, including required
        components such as devices, fluorescence data, imaging planes, and two-photon series. It also applies
        temporary schema adjustments to handle certain use cases until a centralized metadata schema definition
        is available.

        Returns
        -------
        dict
            A dictionary representing the updated Ophys metadata schema.

        Notes
        -----
        - Ensures that `Device` and `ImageSegmentation` are marked as required.
        - Updates various properties, including ensuring arrays for `ImagingPlane` and `TwoPhotonSeries`.
        - Adjusts the schema for `Fluorescence`, including required fields and pattern properties.
        - Adds schema definitions for `DfOverF`, segmentation images, and summary images.
        - Applies temporary fixes, such as setting additional properties for `ImageSegmentation` to True.
        """
        metadata_schema = super().get_metadata_schema()
        metadata_schema["required"] = ["Ophys"]
        metadata_schema["properties"]["Ophys"] = get_base_schema()
        metadata_schema["properties"]["Ophys"]["properties"] = dict(
            Device=dict(type="array", minItems=1, items=get_schema_from_hdmf_class(Device)),
        )
        metadata_schema["properties"]["Ophys"]["properties"].update(
            Fluorescence=get_schema_from_hdmf_class(Fluorescence),
            ImageSegmentation=get_schema_from_hdmf_class(ImageSegmentation),
            ImagingPlane=get_schema_from_hdmf_class(ImagingPlane),
            TwoPhotonSeries=get_schema_from_hdmf_class(TwoPhotonSeries),
        )
        metadata_schema["properties"]["Ophys"]["required"] = ["Device", "ImageSegmentation"]

        # Temporary fixes until centralized definition of metadata schemas
        metadata_schema["properties"]["Ophys"]["properties"]["ImagingPlane"].update(type="array")
        metadata_schema["properties"]["Ophys"]["properties"]["TwoPhotonSeries"].update(type="array")

        metadata_schema["properties"]["Ophys"]["properties"]["Fluorescence"].update(required=["name"])
        metadata_schema["properties"]["Ophys"]["properties"]["Fluorescence"].pop("additionalProperties")

        roi_response_series_schema = metadata_schema["properties"]["Ophys"]["properties"]["Fluorescence"][
            "properties"
        ].pop("roi_response_series")

        roi_response_series_schema.pop("maxItems")
        roi_response_series_schema["items"].update(required=list())

        roi_response_series_per_plane_schema = dict(
            type="object", patternProperties={"^[a-zA-Z0-9]+$": roi_response_series_schema["items"]}
        )

        metadata_schema["properties"]["Ophys"]["properties"]["Fluorescence"].update(
            patternProperties={"^(?!name$)[a-zA-Z0-9]+$": roi_response_series_per_plane_schema}
        )

        metadata_schema["properties"]["Ophys"]["properties"]["ImageSegmentation"]["additionalProperties"] = True

        metadata_schema["properties"]["Ophys"]["properties"]["DfOverF"] = metadata_schema["properties"]["Ophys"][
            "properties"
        ]["Fluorescence"]

        # NOTE: Would prefer to remove in favor of simply using the up-to-date metadata_schema.json
        images_inner_schema = dict(
            type="object",
            properties=dict(name=dict(type="string"), description=dict(type="string")),
        )

        summary_images_per_plane_schema = dict(type="object", patternProperties={"^[a-zA-Z0-9]+$": images_inner_schema})

        metadata_schema["properties"]["Ophys"]["properties"]["SegmentationImages"] = dict(
            type="object",
            required=["name"],
            properties=dict(
                name=dict(type="string", default="SegmentationImages"),
                description=dict(type="string"),
            ),
            patternProperties={
                "^(?!(name|description)$)[a-zA-Z0-9]+$": summary_images_per_plane_schema,
            },
        )

        fill_defaults(metadata_schema, self.get_metadata())
        return metadata_schema

    def get_metadata(self) -> DeepDict:
        from ...tools.roiextractors import get_nwb_segmentation_metadata

        metadata = super().get_metadata()
        metadata.update(get_nwb_segmentation_metadata(self.segmentation_extractor))
        return metadata

    def get_original_timestamps(self) -> np.ndarray:
        reinitialized_extractor = self.get_extractor()(**self.source_data)
        return reinitialized_extractor.get_timestamps()

    def get_timestamps(self) -> np.ndarray:
        return self.segmentation_extractor.get_timestamps()

    def set_aligned_timestamps(self, aligned_timestamps: np.ndarray):
        self.segmentation_extractor.set_times(times=aligned_timestamps)

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata: dict | None = None,
        stub_test: bool = False,
        stub_frames: int | None = None,
        include_background_segmentation: bool = False,
        include_roi_centroids: bool = True,
        include_roi_acceptance: bool = True,
        mask_type: Literal["image", "pixel", "voxel"] = "image",
        plane_segmentation_name: str | None = None,
        iterator_options: dict | None = None,
        stub_samples: int = 100,
    ):
        """
        Add segmentation data to the NWB file.

        Parameters
        ----------
        nwbfile : NWBFile
            The NWBFile to add the plane segmentation to.
        metadata : dict, optional
            The metadata for the interface
        stub_test : bool, default: False
        stub_frames : int, optional
            .. deprecated:: February 2026
                Use `stub_samples` instead.
        include_background_segmentation : bool, default: False
            Whether to include the background plane segmentation and fluorescence traces in the NWB file. If False,
            neuropil traces are included in the main plane segmentation rather than the background plane segmentation.
        include_roi_centroids : bool, default: True
            Whether to include the ROI centroids on the PlaneSegmentation table.
            If there are a very large number of ROIs (such as in whole-brain recordings),
            you may wish to disable this for faster write speeds.
        include_roi_acceptance : bool, default: True
            Whether to include if the detected ROI was 'accepted' or 'rejected'.
            If there are a very large number of ROIs (such as in whole-brain recordings), you may wish to disable this for
            faster write speeds.
        mask_type : str, default: 'image'
            There are three types of ROI masks in NWB, 'image', 'pixel', and 'voxel'.

            * 'image' masks have the same shape as the reference images the segmentation was applied to, and weight each pixel
              by its contribution to the ROI (typically boolean, with 0 meaning 'not in the ROI').
            * 'pixel' masks are instead indexed by ROI, with the data at each index being the shape of the image by the number
              of pixels in each ROI.
            * 'voxel' masks are instead indexed by ROI, with the data at each index being the shape of the volume by the number
              of voxels in each ROI.

            Specify your choice between these two as mask_type='image', 'pixel', 'voxel'
        plane_segmentation_name : str, optional
            The name of the plane segmentation to be added.
        iterator_options : dict, optional
            Options for controlling the iterative write process (buffer size, progress bars) when
            writing image masks and traces.

            Note: To configure chunk size and compression, use the backend configuration system
            via ``get_default_backend_configuration()`` and ``configure_backend()`` after calling
            this method. See the backend configuration documentation for details.
        stub_samples : int, default: 100
            The number of samples (frames) to use for testing. When provided, takes precedence over `stub_frames`.

        Returns
        -------

        """
        from ...tools.roiextractors import add_segmentation_to_nwbfile

        # Handle deprecation of stub_frames in favor of stub_samples
        if stub_frames is not None and stub_samples != 100:
            raise ValueError("Cannot specify both 'stub_frames' and 'stub_samples'. Use 'stub_samples' only.")

        if stub_frames is not None:
            warnings.warn(
                "The 'stub_frames' parameter is deprecated and will be removed on or after February 2026. "
                "Use 'stub_samples' instead.",
                FutureWarning,
                stacklevel=2,
            )
            effective_stub_samples = stub_frames
        else:
            effective_stub_samples = stub_samples

        if stub_test:
            effective_stub_samples = min([effective_stub_samples, self.segmentation_extractor.get_num_samples()])
            segmentation_extractor = self.segmentation_extractor.slice_samples(
                start_sample=0, end_sample=effective_stub_samples
            )
        else:
            segmentation_extractor = self.segmentation_extractor

        metadata = metadata or self.get_metadata()

        add_segmentation_to_nwbfile(
            segmentation_extractor=segmentation_extractor,
            nwbfile=nwbfile,
            metadata=metadata,
            include_background_segmentation=include_background_segmentation,
            include_roi_centroids=include_roi_centroids,
            include_roi_acceptance=include_roi_acceptance,
            mask_type=mask_type,
            plane_segmentation_name=plane_segmentation_name,
            iterator_options=iterator_options,
        )



import uuid
import warnings
from copy import deepcopy
from pathlib import Path

import neo.io.baseio
import numpy as np
import pynwb
from pydantic import FilePath

from ..nwb_helpers import add_device_from_metadata

response_classes = dict(
    voltage_clamp=pynwb.icephys.VoltageClampSeries,
    current_clamp=pynwb.icephys.CurrentClampSeries,
    izero=pynwb.icephys.IZeroClampSeries,
)

stim_classes = dict(
    voltage_clamp=pynwb.icephys.VoltageClampStimulusSeries,
    current_clamp=pynwb.icephys.CurrentClampStimulusSeries,
)


# TODO - get electrodes metadata
def get_electrodes_metadata(neo_reader, electrodes_ids: list, block: int = 0) -> list:
    """
    Get electrodes metadata from Neo reader.

    The typical information we look for is the information accepted by pynwb.icephys.IntracellularElectrode:
      - name – the name of this electrode
      - device – the device that was used to record from this electrode
      - description – Recording description, description of electrode (e.g., whole-cell, sharp, etc.)
      - comment: Free-form text (can be from Methods)
      - slice – Information about slice used for recording.
      - seal – Information about seal used for recording.
      - location – Area, layer, comments on estimation, stereotaxis coordinates (if in vivo, etc.).
      - resistance – Electrode resistance COMMENT: unit: Ohm.
      - filtering – Electrode specific filtering.
      - initial_access_resistance – Initial access resistance.

    Parameters
    ----------
    neo_reader ([type]): Neo reader
    electrodes_ids (list): List of electrodes ids.
    block (int, optional): Block id. Defaults to 0.

    Returns
    -------
    list: List of dictionaries containing electrodes metadata.
    """
    return []


def get_number_of_electrodes(neo_reader) -> int:
    """
    Get number of electrodes from Neo reader.

    Returns
    -------
    int
        The total number of electrodes in the recording.
    """
    # TODO - take in account the case with multiple streams.
    return len(neo_reader.header["signal_channels"])


def get_number_of_segments(neo_reader, block: int = 0) -> int:
    """
    Get number of segments from Neo reader.

    Parameters
    ----------
    neo_reader : neo.io.baseio
        The Neo reader object.
    block : int, default: 0
        Block index.

    Returns
    -------
    int
        The number of segments in the specified block.
    """
    return neo_reader.header["nb_segment"][block]


def get_command_traces(neo_reader, segment: int = 0, cmd_channel: int = 0) -> tuple[list, str, str]:
    """
    Get command traces (e.g. voltage clamp command traces).

    Parameters
    ----------
    neo_reader : neo.io.baseio
        The Neo reader object.
    segment : int, optional
        Segment index. Defaults to 0.
    cmd_channel : int, optional
        ABF command channel (0 to 7). Defaults to 0.

    Returns
    -------
    tuple[list, str, str]
        A tuple containing:
        - list: The command trace data
        - str: The title of the command trace
        - str: The units of the command trace

    Notes
    -----
    This function only works for AxonIO interface.
    """
    try:
        traces, titles, units = neo_reader.read_raw_protocol()
        return traces[segment][cmd_channel], titles[segment][cmd_channel], units[segment][cmd_channel]
    except Exception as e:
        msg = ".\n\n WARNING - get_command_traces() only works for AxonIO interface."
        e.args = (str(e) + msg,)
        return e


def get_conversion_from_unit(unit: str) -> float:
    """
    Get conversion (to Volt or Ampere) from unit in string format.

    Parameters
    ----------
    unit : str
        Unit as string. E.g. pA, mV, uV, etc...

    Returns
    -------
    float
        The conversion factor to convert to Ampere or Volt.
        For example, for 'pA' returns 1e-12 to convert to Ampere.
    """
    if unit in ["pA", "pV"]:
        conversion = 1e-12
    elif unit in ["nA", "nV"]:
        conversion = 1e-9
    elif unit in ["uA", "uV"]:
        conversion = 1e-6
    elif unit in ["mA", "mV"]:
        conversion = 1e-3
    elif unit in ["A", "V"]:
        conversion = 1.0
    else:
        conversion = 1.0
        warnings.warn("No valid units found for traces in the current file. Gain is set to 1, but this might be wrong.")
    return float(conversion)


def get_nwb_metadata(neo_reader, metadata: dict = None) -> dict:
    """
    Return default metadata for all recording fields.

    Parameters
    ----------
    neo_reader : neo.io.baseio
        Neo reader object
    metadata : dict, optional
        Metadata info for constructing the nwb file.

    Returns
    -------
    dict
        Default metadata dictionary containing NWBFile and Icephys device information.
    """
    metadata = dict(
        NWBFile=dict(
            session_description="Auto-generated by NwbRecordingExtractor without description.",
            identifier=str(uuid.uuid4()),
        ),
        Icephys=dict(Device=[dict(name="Device", description="no description")]),
    )
    return metadata


def add_icephys_electrode(neo_reader, nwbfile, metadata: dict = None):
    """
    Add icephys electrodes to nwbfile object.

    Will always ensure nwbfile has at least one icephys electrode.
    Will auto-generate a linked device if the specified name does not exist in the nwbfile.

    Parameters
    ----------
    neo_reader : neo.io.baseio
    nwbfile : NWBFile
        NWBFile object to add the icephys electrode to.
    metadata : dict, optional
        Metadata info for constructing the nwb file.
        Should be of the format::

            metadata['Icephys']['Electrodes'] = [
                {
                    'name': my_name,
                    'description': my_description,
                    'device_name': my_device_name
                },
                ...
            ]
    """
    metadata_copy = deepcopy(metadata) if metadata is not None else dict()

    assert isinstance(nwbfile, pynwb.NWBFile), "'nwbfile' should be of type pynwb.NWBFile"

    if len(nwbfile.devices) == 0:
        warnings.warn("When adding Icephys Electrode, no Devices were found on nwbfile. Creating a Device now...")
        add_device_from_metadata(nwbfile=nwbfile, modality="Icephys", metadata=metadata_copy)

    if "Icephys" not in metadata_copy:
        metadata_copy["Icephys"] = dict()

    defaults = [
        dict(
            name=f"icephys_electrode_{elec_id}",
            description="no description",
            device_name=[i.name for i in nwbfile.devices.values()][0],
        )
        for elec_id in range(get_number_of_electrodes(neo_reader))
    ]

    if "Electrodes" not in metadata_copy["Icephys"] or len(metadata_copy["Icephys"]["Electrodes"]) == 0:
        metadata_copy["Icephys"]["Electrodes"] = defaults

    assert all(
        [isinstance(x, dict) for x in metadata_copy["Icephys"]["Electrodes"]]
    ), "Expected metadata['Icephys']['Electrodes'] to be a list of dictionaries!"

    # Create Icephys electrode from metadata
    for elec in metadata_copy["Icephys"]["Electrodes"]:
        if elec.get("name", defaults[0]["name"]) not in nwbfile.icephys_electrodes:
            device_name = elec.pop("device_name", None) or elec.pop("device", defaults[0]["device_name"])
            # elec.pop("device_name", 0)
            if device_name not in nwbfile.devices:
                new_device_metadata = dict(Ecephys=dict(Device=[dict(name=device_name)]))
                add_device_from_metadata(nwbfile, modality="Icephys", metadata=new_device_metadata)
                warnings.warn(
                    f"Device '{device_name}' not detected in "
                    "attempted link to icephys electrode! Automatically generating."
                )
            electrode_kwargs = elec
            electrode_kwargs.update(
                name=elec.get("name", defaults[0]["name"]),
                description=elec.get("description", defaults[0]["description"]),
                device=nwbfile.devices[device_name],
            )
            nwbfile.create_icephys_electrode(**electrode_kwargs)


def add_icephys_recordings(
    neo_reader,
    nwbfile: pynwb.NWBFile,
    metadata: dict = None,
    icephys_experiment_type: str = "voltage_clamp",
    stimulus_type: str = "not described",
    skip_electrodes: tuple[int] = (),
):
    """
    Add icephys recordings (stimulus/response pairs) to nwbfile object.

    Parameters
    ----------
    neo_reader : neo.io.baseio
    nwbfile : NWBFile
    metadata : dict, optional
    icephys_experiment_type : {'voltage_clamp', 'current_clamp', 'izero'}
        Type of icephys recording.
    stimulus_type : str, default: 'not described'
    skip_electrodes : tuple, default: ()
        Electrode IDs to skip.
    """

    n_segments = get_number_of_segments(neo_reader, block=0)

    # Check for protocol data (only ABF2), necessary for stimuli data
    if neo_reader._axon_info["fFileVersionNumber"] < 2:
        n_commands = 0
        warnings.warn(
            f"Protocol section is only present in ABF2 files. {neo_reader.filename} has version "
            f"{neo_reader._axon_info['fFileVersionNumber']}. Saving experiment as 'i_zero'..."
        )
    else:
        protocol = neo_reader.read_raw_protocol()
        n_commands = len(protocol[0])

    if n_commands == 0:
        icephys_experiment_type = "izero"
        warnings.warn(
            f"No command data found by neo reader in file {neo_reader.filename}. Saving experiment as 'i_zero'..."
        )
    else:
        assert (
            n_commands == n_segments
        ), f"File contains inconsistent number of segments ({n_segments}) and commands ({n_commands})"

    assert icephys_experiment_type in ["voltage_clamp", "current_clamp", "izero"], (
        f"'icephys_experiment_type' should be 'voltage_clamp', 'current_clamp' or 'izero', but received value "
        f"{icephys_experiment_type}"
    )

    # Check and auto-create electrodes, in case they don't exist yet in nwbfile
    if len(nwbfile.icephys_electrodes) == 0:
        warnings.warn(
            "When adding Icephys Recording, no Icephys Electrodes were found on nwbfile. Creating Electrodes now..."
        )
        add_icephys_electrode(
            neo_reader=neo_reader,
            nwbfile=nwbfile,
            metadata=metadata,
        )

    if getattr(nwbfile, "intracellular_recordings", None):
        ri = max(nwbfile.intracellular_recordings["responses"].index)
    else:
        ri = -1

    if getattr(nwbfile, "icephys_simultaneous_recordings", None):
        simultaneous_recordings_offset = len(nwbfile.icephys_simultaneous_recordings)
    else:
        simultaneous_recordings_offset = 0

    if getattr(nwbfile, "icephys_sequential_recordings", None):
        sessions_offset = len(nwbfile.icephys_sequential_recordings)
    else:
        sessions_offset = 0

    relative_session_start_time = deepcopy(
        metadata["Icephys"]["Sessions"][sessions_offset]["relative_session_start_time"]
    )
    session_stimulus_type = deepcopy(metadata["Icephys"]["Sessions"][sessions_offset]["stimulus_type"])

    # Sequential icephys recordings
    simultaneous_recordings = list()
    for si in range(n_segments):
        # Parallel icephys recordings
        recordings = list()
        for ei, electrode in enumerate(
            list(nwbfile.icephys_electrodes.values())[: len(neo_reader.header["signal_channels"]["units"])]
        ):
            if ei in skip_electrodes:
                continue
            # Starting time is the signal starting time within .abf file + time
            # relative to first session (first .abf file)
            ri += 1
            starting_time = neo_reader.get_signal_t_start(block_index=0, seg_index=si)
            starting_time = starting_time + relative_session_start_time

            sampling_rate = neo_reader.get_signal_sampling_rate()
            response_unit = neo_reader.header["signal_channels"]["units"][ei]
            response_conversion = get_conversion_from_unit(unit=response_unit)
            response_gain = neo_reader.header["signal_channels"]["gain"][ei]
            response_name = f"{icephys_experiment_type}-response-{si + 1 + simultaneous_recordings_offset:02}-ch-{ei}"

            response = response_classes[icephys_experiment_type](
                name=response_name,
                description=f"Response to: {session_stimulus_type}",
                electrode=electrode,
                data=neo_reader.get_analogsignal_chunk(block_index=0, seg_index=si, channel_indexes=ei),
                starting_time=starting_time,
                rate=sampling_rate,
                conversion=response_conversion * response_gain,
                gain=np.nan,
            )
            if icephys_experiment_type != "izero":
                stim_unit = protocol[2][ei]
                stim_conversion = get_conversion_from_unit(unit=stim_unit)
                stimulus = stim_classes[icephys_experiment_type](
                    name=f"stimulus-{si + 1 + simultaneous_recordings_offset:02}-ch-{ei}",
                    description=f"Stim type: {session_stimulus_type}",
                    electrode=electrode,
                    data=protocol[0][si][ei],
                    rate=sampling_rate,
                    starting_time=starting_time,
                    conversion=stim_conversion,
                    gain=np.nan,
                )
                icephys_recording = nwbfile.add_intracellular_recording(
                    electrode=electrode, response=response, stimulus=stimulus
                )
            else:
                icephys_recording = nwbfile.add_intracellular_recording(electrode=electrode, response=response)

            recordings.append(icephys_recording)

        sim_rec = nwbfile.add_icephys_simultaneous_recording(recordings=recordings)
        simultaneous_recordings.append(sim_rec)

    nwbfile.add_icephys_sequential_recording(
        simultaneous_recordings=simultaneous_recordings, stimulus_type=stimulus_type
    )

    # TODO
    # # Add a list of sequential recordings table indices as a repetition
    # run_index = nwbfile.add_icephys_repetition(
    #     sequential_recordings=[
    #         seq_rec,
    #     ]
    # )

    # # Add a list of repetition table indices as a experimental condition
    # nwbfile.add_icephys_experimental_condition(
    #     repetitions=[
    #         run_index,
    #     ]
    # )


def add_neo_to_nwb(
    neo_reader,
    nwbfile: pynwb.NWBFile,
    metadata: dict = None,
    icephys_experiment_type: str = "voltage_clamp",
    stimulus_type: str | None = None,
    skip_electrodes: tuple[int] = (),
):
    """
    Auxiliary static method for nwbextractor.

    Adds all recording related information from recording object and metadata to the nwbfile object.

    Parameters
    ----------
    neo_reader: Neo reader object
    nwbfile: NWBFile
        nwb file to which the recording information is to be added
    metadata: dict
        metadata info for constructing the nwb file (optional).
        Check the auxiliary function docstrings for more information
        about metadata format.
    icephys_experiment_type: str (optional)
        Type of Icephys experiment. Allowed types are: 'voltage_clamp', 'current_clamp' and 'izero'.
        If no value is passed, 'voltage_clamp' is used as default.
    stimulus_type: str, optional
    skip_electrodes: tuple, optional
        Electrode IDs to skip. Defaults to ().
    """
    assert isinstance(nwbfile, pynwb.NWBFile), "'nwbfile' should be of type pynwb.NWBFile"

    # TODO: remove completely after 10/1/2024

    add_device_from_metadata(nwbfile=nwbfile, modality="Icephys", metadata=metadata)

    add_icephys_electrode(
        neo_reader=neo_reader,
        nwbfile=nwbfile,
        metadata=metadata,
    )

    add_icephys_recordings(
        neo_reader=neo_reader,
        nwbfile=nwbfile,
        metadata=metadata,
        icephys_experiment_type=icephys_experiment_type,
        stimulus_type=stimulus_type,
        skip_electrodes=skip_electrodes,
    )


def write_neo_to_nwb(
    neo_reader: neo.io.baseio.BaseIO,
    save_path: FilePath | None = None,  # pragma: no cover
    overwrite: bool = False,
    nwbfile=None,
    metadata: dict = None,
    icephys_experiment_type: str | None = None,
    stimulus_type: str | None = None,
    skip_electrodes: tuple | None = (),
):
    """
    Primary method for writing a Neo reader object to an NWBFile.

    Parameters
    ----------
    neo_reader: Neo reader
    save_path: PathType
        Required if an nwbfile is not passed. Must be the path to the nwbfile
        being appended, otherwise one is created and written.
    overwrite: bool
        If using save_path, whether to overwrite the NWBFile if it already exists.
    nwbfile: NWBFile
        Required if a save_path is not specified. If passed, this function
        will fill the relevant fields within the nwbfile.
    metadata: dict
        metadata info for constructing the nwb file (optional). Should be of the format::

            metadata['Ecephys'] = {}

        with keys of the forms::

            metadata['Ecephys']['Device'] = [
                {
                    'name': my_name,
                    'description': my_description
                },
                ...
            ]
            metadata['Ecephys']['ElectrodeGroup'] = [
                {
                    'name': my_name,
                    'description': my_description,
                    'location': electrode_location,
                    'device': my_device_name
                },
                ...
            ]
            metadata['Ecephys']['Electrodes'] = [
                {
                    'name': my_name,
                    'description': my_description
                },
                ...
            ]
            metadata['Ecephys']['ElectricalSeries'] = {
                'name': my_name,
                'description': my_description
            }

        Note that data intended to be added to the electrodes table of the NWBFile should be set as channel
        properties in the RecordingExtractor object.
    icephys_experiment_type: str (optional)
        Type of Icephys experiment. Allowed types are: 'voltage_clamp', 'current_clamp' and 'izero'.
        If no value is passed, 'voltage_clamp' is used as default.
    stimulus_type: str, optional
    skip_electrodes: tuple, optional
        Electrode IDs to skip. Defaults to ().
    """
    if nwbfile is not None:
        assert isinstance(nwbfile, pynwb.NWBFile), "'nwbfile' should be of type pynwb.NWBFile"

    assert save_path is None or nwbfile is None, "Either pass a save_path location, or nwbfile object, but not both!"

    if metadata is None:
        metadata = get_nwb_metadata(neo_reader=neo_reader)

    kwargs = dict(
        neo_reader=neo_reader,
        metadata=metadata,
        icephys_experiment_type=icephys_experiment_type,
        stimulus_type=stimulus_type,
        skip_electrodes=skip_electrodes,
    )
    if nwbfile is None:
        if Path(save_path).is_file() and not overwrite:
            read_mode = "r+"
        else:
            read_mode = "w"

        with pynwb.NWBHDF5IO(str(save_path), mode=read_mode) as io:
            if read_mode == "r+":
                nwbfile = io.read()
            else:
                nwbfile_kwargs = dict(
                    session_description="Auto-generated by NwbRecordingExtractor without description.",
                    identifier=str(uuid.uuid4()),
                )
                if metadata is not None and "NWBFile" in metadata:
                    nwbfile_kwargs.update(metadata["NWBFile"])
                nwbfile = pynwb.NWBFile(**nwbfile_kwargs)

            add_neo_to_nwb(nwbfile=nwbfile, **kwargs)
            io.write(nwbfile)
    else:
        add_neo_to_nwb(nwbfile=nwbfile, **kwargs)



import warnings
from pathlib import Path

import numpy as np
import pandas as pd
from pydantic import FilePath, validate_call
from pynwb.file import NWBFile

from ....basetemporalalignmentinterface import BaseTemporalAlignmentInterface
from ....utils import DeepDict


class DeepLabCutInterface(BaseTemporalAlignmentInterface):
    """Data interface for DeepLabCut datasets."""

    display_name = "DeepLabCut"
    keywords = ("DLC", "DeepLabCut", "pose estimation", "behavior")
    associated_suffixes = (".h5", ".csv")
    info = "Interface for handling data from DeepLabCut."

    _timestamps = None

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = super().get_source_schema()
        source_schema["properties"]["file_path"]["description"] = "Path to the file output by dlc (.h5 or .csv)."
        source_schema["properties"]["config_file_path"]["description"] = "Path to .yml config file."
        return source_schema

    def get_available_subjects(file_path: FilePath) -> list[str]:
        """
        Extract available subjects from a DeepLabCut output file.

        Parameters
        ----------
        file_path : FilePath
            Path to the DeepLabCut output file (.h5 or .csv).

        Returns
        -------
        list[str]
            List of subject names found in the file.

        Raises
        ------
        IOError
            If the file is not a valid DeepLabCut output file.
        FileNotFoundError
            If the file does not exist.
        """
        file_path = Path(file_path)

        if not file_path.exists():
            raise FileNotFoundError(f"File {file_path} does not exist.")

        # Read the data
        if ".h5" in file_path.suffixes:
            df = pd.read_hdf(file_path)
        elif ".csv" in file_path.suffixes:
            df = pd.read_csv(file_path, header=[0, 1, 2], index_col=0)
        else:
            raise IOError(f"The file {file_path} passed in is not a valid DeepLabCut output data file.")

        # Check if 'individuals' level exists in the column structure
        if "individuals" in df.columns.names:
            # Multi-subject file - extract unique individuals
            individuals = df.columns.get_level_values("individuals").unique().tolist()
            return individuals
        else:
            # Single-subject file - return default subject name
            # For consistency with the interface's default behavior
            return ["ind1"]

    @validate_call
    def __init__(
        self,
        file_path: FilePath,
        config_file_path: FilePath | None = None,
        subject_name: str = "ind1",
        pose_estimation_metadata_key: str = "PoseEstimationDeepLabCut",
        verbose: bool = False,
    ):
        """
        Interface for writing DeepLabCut's output files to NWB.

        This interface reads DeepLabCut output files (.h5 or .csv) and converts them to NWB format
        using the ndx-pose extension. It extracts keypoints (bodyparts), their coordinates, and confidence
        values, and organizes them into a structured format within the NWB file.

        Parameters
        ----------
        file_path : FilePath
            Path to the file output by DeepLabCut (.h5 or .csv). The file should contain the pose estimation
            data with keypoints, coordinates, and confidence values.
        config_file_path : FilePath, optional
            Path to the DeepLabCut .yml config file. If provided, additional metadata such as video dimensions,
            task description, and experimenter information will be extracted.
        subject_name : str, default: "ind1"
            The subject name to be used in the metadata. For output files with multiple individuals,
            this must match the name of the individual for which the data will be added. This name is also
            used to link the skeleton to the subject in the NWB file.
        pose_estimation_metadata_key : str, default: "PoseEstimationDeepLabCut"
            This controls where in the metadata the pose estimation metadata is stored:
            metadata["PoseEstimation"]["PoseEstimationContainers"][pose_estimation_metadata_key].
            This key is also used as the name of the PoseEstimation container in the NWB file.
        verbose : bool, default: False
            Controls verbosity of the conversion process.


        Metadata Structure
        ------------------
        The metadata is organized in a hierarchical structure:

        .. code-block:: python

            metadata = {
                "PoseEstimation": {
                    "Skeletons": {
                        "skeleton_name": {
                            "name": "SkeletonPoseEstimationDeepLabCut_SubjectName",
                            "nodes": ["bodypart1", "bodypart2", ...],  # List of keypoints/bodyparts
                            "edges": [[0, 1], [1, 2], ...],  # Connections between nodes (optional)
                            "subject": "subject_name"  # Links the skeleton to the subject
                        }
                    },
                    "Devices": {
                        "device_name": {
                            "name": "CameraPoseEstimationDeepLabCut",
                            "description": "Camera used for behavioral recording and pose estimation."
                        }
                    },
                    "PoseEstimationContainers": {
                        "pose_estimation_metadata_key": {
                            "name": "PoseEstimationDeepLabCut",
                            "description": "2D keypoint coordinates estimated using DeepLabCut.",
                            "source_software": "DeepLabCut",
                            "devices": ["device_name"],  # References to devices
                            "PoseEstimationSeries": {
                                "PoseEstimationSeriesBodyPart1": {
                                    "name": "bodypart1",
                                    "description": "Keypoint bodypart1.",
                                    "unit": "pixels",
                                    "reference_frame": "(0,0) corresponds to the bottom left corner of the video.",
                                    "confidence_definition": "Softmax output of the deep neural network."
                                },
                                "PoseEstimationSeriesBodyPart2": {
                                    "name": "bodypart2",
                                    "description": "Keypoint bodypart2.",
                                    "unit": "pixels",
                                    "reference_frame": "(0,0) corresponds to the bottom left corner of the video.",
                                    "confidence_definition": "Softmax output of the deep neural network."
                                }
                                # And so on for each bodypart
                            }
                        }
                    }
                }
            }

        The metadata can be customized by:

        #. Calling get_metadata() to retrieve the default metadata
        #. Modifying the returned dictionary as needed
        #. Passing the modified metadata to add_to_nwbfile() or run_conversion()

        See also our `Conversion Gallery <https://neuroconv.readthedocs.io/en/main/conversion_examples_gallery/behavior/deeplabcut.html>`_
        for more examples using DeepLabCut data.

        Notes
        -----
        - When the subject_name matches a subject_id in the NWBFile, the skeleton will be automatically
        linked to that subject.
        """
        # This import is to assure that the ndx_pose is in the global namespace when an pynwb.io object is created
        from importlib.metadata import version

        import ndx_pose  # noqa: F401
        from packaging import version as version_parse

        ndx_pose_version = version("ndx-pose")
        if version_parse.parse(ndx_pose_version) < version_parse.parse("0.2.0"):
            raise ImportError(
                "DeepLabCut interface requires ndx-pose version 0.2.0 or later. "
                f"Found version {ndx_pose_version}. Please upgrade: "
                "pip install 'ndx-pose>=0.2.0'"
            )

        from ._dlc_utils import _read_config

        file_path = Path(file_path)
        suffix_is_valid = ".h5" in file_path.suffixes or ".csv" in file_path.suffixes
        if not "DLC" in file_path.stem or not suffix_is_valid:
            raise IOError("The file passed in is not a valid DeepLabCut output data file.")

        self.config_dict = dict()
        if config_file_path is not None:
            self.config_dict = _read_config(config_file_path=config_file_path)
        self.subject_name = subject_name
        self.verbose = verbose
        self.pose_estimation_metadata_key = pose_estimation_metadata_key
        self.pose_estimation_container_kwargs = dict()

        super().__init__(file_path=file_path, config_file_path=config_file_path)

    def get_metadata_schema(self) -> dict:
        """
        Retrieve JSON schema for metadata specific to the DeepLabCutInterface.

        Returns
        -------
        dict
            The JSON schema defining the metadata structure.
        """
        from ....utils import get_base_schema

        metadata_schema = super().get_metadata_schema()

        # Define the schema for PoseEstimation metadata
        metadata_schema["properties"]["PoseEstimation"] = get_base_schema(tag="PoseEstimation")

        # Add Skeletons schema
        skeleton_schema = get_base_schema(tag="Skeletons")
        skeleton_schema["additionalProperties"] = {
            "type": "object",
            "properties": {
                "name": {"type": "string", "description": "Name of the skeleton"},
                "nodes": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "List of node names (bodyparts)",
                },
                "edges": {
                    "type": ["array", "null"],
                    "items": {
                        "type": "array",
                        "items": {"type": "integer"},
                        "minItems": 2,
                        "maxItems": 2,
                    },
                    "description": "List of edges connecting nodes, each edge is a pair of node indices",
                },
                "subject": {
                    "type": ["string", "null"],
                    "description": "Subject ID associated with this skeleton",
                },
            },
            "required": ["name", "nodes"],
        }

        # Add Devices schema
        devices_schema = get_base_schema(tag="Devices")
        devices_schema["additionalProperties"] = {
            "type": "object",
            "properties": {
                "name": {
                    "type": "string",
                    "description": "Name of the device",
                },
                "description": {
                    "type": "string",
                    "description": "Description of the device",
                },
            },
            "required": ["name"],
        }

        # Add PoseEstimationContainers schema
        containers_schema = get_base_schema(tag="PoseEstimationContainers")
        containers_schema["additionalProperties"] = {
            "type": "object",
            "description": "Metadata for a PoseEstimation group corresponding to one subject/session",
            "properties": {
                "name": {
                    "type": "string",
                    "description": "Name of the PoseEstimation group",
                    "default": "PoseEstimationDeepLabCut",
                },
                "description": {
                    "type": ["string", "null"],
                    "description": "Description of the pose estimation procedure and output",
                },
                "source_software": {
                    "type": ["string", "null"],
                    "description": "Name of the software tool used",
                    "default": "DeepLabCut",
                },
                "source_software_version": {
                    "type": ["string", "null"],
                    "description": "Version string of the software tool used",
                },
                "scorer": {
                    "type": ["string", "null"],
                    "description": "Name of the scorer or algorithm used",
                },
                "dimensions": {
                    "type": ["array", "null"],
                    "description": "Dimensions [height, width] of the labeled video(s)",
                    "items": {"type": "array", "items": {"type": "integer"}},
                },
                "original_videos": {
                    "type": ["array", "null"],
                    "description": "Paths to the original video files",
                    "items": {"type": "string"},
                },
                "labeled_videos": {
                    "type": ["array", "null"],
                    "description": "Paths to the labeled video files",
                    "items": {"type": "string"},
                },
                "skeleton": {
                    "type": ["string", "null"],
                    "description": "Reference to a Skeleton defined in Skeletons",
                },
                "devices": {
                    "type": ["array", "null"],
                    "description": "References to Device objects used to record the videos",
                    "items": {"type": "string"},
                },
                "PoseEstimationSeries": {
                    "type": ["object", "null"],
                    "description": "Dictionary of PoseEstimationSeries, one per body part",
                    "additionalProperties": {
                        "type": "object",
                        "properties": {
                            "name": {
                                "type": ["string", "null"],
                                "description": "Name for this series, typically the body part",
                            },
                            "description": {
                                "type": ["string", "null"],
                                "description": "Description for this specific series",
                            },
                            "unit": {
                                "type": ["string", "null"],
                                "description": "Unit of measurement (default: pixels)",
                                "default": "pixels",
                            },
                            "reference_frame": {
                                "type": ["string", "null"],
                                "description": "Description of the reference frame",
                                "default": "(0,0) corresponds to the bottom left corner of the video.",
                            },
                            "confidence_definition": {
                                "type": ["string", "null"],
                                "description": "How the confidence was computed (e.g., Softmax output)",
                                "default": "Softmax output of the deep neural network.",
                            },
                        },
                        "required": ["name"],
                    },
                },
            },
            "required": ["name"],
        }

        # Add all schemas to the PoseEstimation schema
        metadata_schema["properties"]["PoseEstimation"]["properties"] = {
            "Skeletons": skeleton_schema,
            "Devices": devices_schema,
            "PoseEstimationContainers": containers_schema,
        }

        return metadata_schema

    def get_metadata(self) -> DeepDict:
        metadata = super().get_metadata()

        if self.config_dict:
            metadata["NWBFile"].update(
                session_description=self.config_dict["Task"],
                experimenter=[self.config_dict["scorer"]],
            )

        # Extract information from the DeepLabCut data
        file_path = self.source_data["file_path"]

        # Read the data to extract bodyparts
        if ".h5" in Path(file_path).suffixes:
            df = pd.read_hdf(file_path)
        elif ".csv" in Path(file_path).suffixes:
            df = pd.read_csv(file_path, header=[0, 1, 2], index_col=0)

        # Ensure individuals in header if needed
        from ._dlc_utils import _ensure_individuals_in_header

        df = _ensure_individuals_in_header(df, self.subject_name)

        # Extract bodyparts and individuals
        bodyparts = df.columns.get_level_values("bodyparts").unique().tolist()
        individuals = df.columns.get_level_values("individuals").unique().tolist()

        # Get video dimensions from config if available
        dimensions = None
        if self.source_data.get("config_file_path"):
            from ._dlc_utils import _get_video_info_from_config_file

            video_name = Path(file_path).stem.split("DLC")[0]
            _, image_shape = _get_video_info_from_config_file(
                config_file_path=self.source_data["config_file_path"], vidname=video_name
            )
            # Parse dimensions from image_shape (format: "0, width, 0, height")
            try:
                shape_parts = [int(x.strip()) for x in image_shape.split(",")]
                if len(shape_parts) == 4:
                    dimensions = [[shape_parts[3], shape_parts[1]]]  # [[height, width]]
            except (ValueError, IndexError):
                pass

        # Get edges from metadata pickle file if available
        edges = []
        try:
            from ._dlc_utils import _get_graph_edges

            filename = str(Path(file_path).parent / Path(file_path).stem)
            for i, c in enumerate(filename[::-1]):
                if c.isnumeric():
                    break
            if i > 0:
                filename = filename[:-i]

            metadata_file_path = Path(filename + "_meta.pickle")
            edges = _get_graph_edges(metadata_file_path=metadata_file_path)
        except Exception:
            pass

        # Create default PoseEstimation metadata
        container_name = self.pose_estimation_metadata_key
        skeleton_name = f"Skeleton{container_name}_{self.subject_name.capitalize()}"
        device_name = f"Camera{container_name}"

        # Create PoseEstimation metadata structure
        pose_estimation_metadata = DeepDict()

        # Add Skeleton as a dictionary
        pose_estimation_metadata["Skeletons"] = {
            skeleton_name: {"name": skeleton_name, "nodes": bodyparts, "edges": edges, "subject": self.subject_name}
        }

        # Add Device as a dictionary
        pose_estimation_metadata["Devices"] = {
            device_name: {
                "name": device_name,
                "description": "Camera used for behavioral recording and pose estimation.",
            }
        }

        # Extract video name and scorer
        video_name, scorer = Path(file_path).stem.split("DLC")
        scorer = "DLC" + scorer

        # Get video info from config file if available
        video_file_path = None
        image_shape = "0, 0, 0, 0"

        if self.source_data.get("config_file_path"):
            from ._dlc_utils import _get_video_info_from_config_file

            video_file_path, image_shape = _get_video_info_from_config_file(
                config_file_path=self.source_data["config_file_path"], vidname=video_name
            )

        # Add PoseEstimation container
        pose_estimation_metadata["PoseEstimationContainers"] = {
            container_name: {
                "name": container_name,
                "description": "2D keypoint coordinates estimated using DeepLabCut.",
                "source_software": "DeepLabCut",
                "dimensions": dimensions,
                "skeleton": skeleton_name,
                "devices": [device_name],
                "scorer": scorer,
                "original_videos": [video_file_path] if video_file_path else None,
                "PoseEstimationSeries": {},
            }
        }

        # Add a series for each bodypart
        for bodypart in bodyparts:
            pose_estimation_metadata["PoseEstimationContainers"][container_name]["PoseEstimationSeries"][bodypart] = {
                "name": f"PoseEstimationSeries{bodypart.capitalize()}",
                "description": f"Pose estimation series for {bodypart}.",
                "unit": "pixels",
                "reference_frame": "(0,0) corresponds to the bottom left corner of the video.",
                "confidence_definition": "Softmax output of the deep neural network.",
            }

        # Add PoseEstimation metadata to the main metadata
        metadata["PoseEstimation"] = pose_estimation_metadata

        return metadata

    def get_original_timestamps(self) -> np.ndarray:
        raise NotImplementedError(
            "Unable to retrieve the original unaltered timestamps for this interface! "
            "Define the `get_original_timestamps` method for this interface."
        )

    def get_timestamps(self) -> np.ndarray:
        raise NotImplementedError(
            "Unable to retrieve timestamps for this interface! Define the `get_timestamps` method for this interface."
        )

    def set_aligned_timestamps(self, aligned_timestamps: list | np.ndarray):
        """
        Set aligned timestamps vector for DLC data with user defined timestamps

        Parameters
        ----------
        aligned_timestamps : list, np.ndarray
            A timestamps vector.
        """
        self._timestamps = np.asarray(aligned_timestamps)

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata: dict | None = None,
        container_name: str | None = None,
    ):
        """
        Conversion from DLC output files to nwb. Derived from dlc2nwb library.

        Parameters
        ----------
        nwbfile: NWBFile
            nwb file to which the recording information is to be added
        metadata: dict
            metadata info for constructing the nwb file (optional).
        container_name: str, default: None
            name of the PoseEstimation container in the nwb. If None, uses the container_name from the interface.
            This parameter is deprecated and will be removed on or after October 2025.
            Use the pose_estimation_metadata_key parameter when initializing the interface instead to specify
            the content of the metadata.

        """
        from ._dlc_utils import (
            _add_pose_estimation_to_nwbfile,
            _ensure_individuals_in_header,
        )

        # Use the pose_estimation_metadata_key from the instance if container_name not provided
        if container_name is not None:
            warnings.warn(
                "The container_name parameter in add_to_nwbfile is deprecated and will be removed on or after October 2025. "
                "Use the pose_estimation_metadata_key parameter when initializing the interface instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            pose_estimation_metadata_key = container_name
        else:
            pose_estimation_metadata_key = self.pose_estimation_metadata_key

        # Get default metadata
        default_metadata = DeepDict(self.get_metadata())

        # Update with user-provided metadata if available
        if metadata is not None:
            default_metadata.deep_update(metadata)

        # Set the container name in the metadata, remove this once container_name is deprecated
        if container_name is not None:
            if (
                "PoseEstimation" in default_metadata
                and "PoseEstimationContainers" in default_metadata["PoseEstimation"]
            ):
                if container_name in default_metadata["PoseEstimation"]["PoseEstimationContainers"]:
                    default_metadata["PoseEstimation"]["PoseEstimationContainers"][container_name][
                        "name"
                    ] = container_name
                else:
                    # If the container doesn't exist in the metadata, create it with the name
                    default_metadata["PoseEstimation"]["PoseEstimationContainers"][container_name] = {
                        "name": container_name
                    }

        file_path = Path(self.source_data["file_path"])

        # Read the data
        if ".h5" in file_path.suffixes:
            df = pd.read_hdf(file_path)
        elif ".csv" in file_path.suffixes:
            df = pd.read_csv(file_path, header=[0, 1, 2], index_col=0)

        # Ensure individuals in header
        df = _ensure_individuals_in_header(df, self.subject_name)

        # Get timestamps
        timestamps = self._timestamps
        if timestamps is None:
            timestamps = df.index.tolist()  # Use index as dummy timestamps if not provided

        df_animal = df.xs(self.subject_name, level="individuals", axis=1)

        _add_pose_estimation_to_nwbfile(
            nwbfile=nwbfile,
            df_animal=df_animal,
            timestamps=timestamps,
            exclude_nans=False,
            metadata=default_metadata,
            pose_estimation_metadata_key=pose_estimation_metadata_key,
        )



from pathlib import Path

from pydantic import FilePath
from pynwb import NWBFile

from ....basedatainterface import BaseDataInterface
from ....utils import DeepDict, get_json_schema_from_method_signature


class EDFAnalogInterface(BaseDataInterface):
    """
    Primary data interface for converting auxiliary data streams from EDF files.

    This interface is designed to handle all the signals that should NOT be stored as ElectricalSeries,
    including physiological monitoring signals, triggers and any other auxiliary data which does not
    come from electrode channels.

    If your data consists of electrical recording channels you should use the
    :py:class:`~neuroconv.datainterfaces.ecephys.edf.edfdatainterface.EDFRecordingInterface`.
    """

    display_name = "EDF Analog"
    keywords = ("edf", "analog", "physiological", "trigger", "auxiliary")
    associated_suffixes = (".edf",)
    info = "Interface for converting EDF analog data (from auxiliary channels)."

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = get_json_schema_from_method_signature(method=cls.__init__)
        source_schema["properties"]["file_path"]["description"] = "Path to the .edf file."
        return source_schema

    @staticmethod
    def get_available_channel_ids(file_path: FilePath) -> list:
        """
        Get all available channel names from an EDF file.

        Parameters
        ----------
        file_path : FilePath
            Path to the EDF file

        Returns
        -------
        list
            List of all channel names in the EDF file
        """
        from spikeinterface.extractors import read_edf

        recording = read_edf(file_path=file_path, all_annotations=True, use_names_as_ids=True)
        channel_ids = recording.get_channel_ids()

        # Clean up to avoid dangling references
        del recording

        return channel_ids.tolist()

    def __init__(
        self,
        *,
        file_path: FilePath,
        channels_to_include: list[str] | None = None,
        verbose: bool = False,
        metadata_key: str = "analog_edf_metadata_key",
    ):
        """
        Load and prepare analog data from EDF format.

        Parameters
        ----------
        file_path : FilePath
            Path to the EDF file
        channels_to_include : list of str, optional
            Specific channel IDs to include.
        verbose : bool, default: False
            Verbose output
        metadata_key : str, default: "analog_edf_metadata_key"
            Key for the TimeSeries metadata in the metadata dictionary.
        """
        from spikeinterface.extractors import read_edf

        self._file_path = Path(file_path)
        self.metadata_key = metadata_key

        full_recording = read_edf(file_path=self._file_path, all_annotations=True, use_names_as_ids=True)

        # Validate that the requested channels exist
        self._channels_to_include = channels_to_include or full_recording.get_channel_ids().tolist()
        available_channels = full_recording.get_channel_ids().astype(str)
        missing_channels = set(self._channels_to_include) - set(available_channels)
        if missing_channels:
            error_msg = (
                f"Channels not found in EDF file: {missing_channels}. "
                f"Available channels: {list(available_channels)}"
            )
            raise ValueError(error_msg)

        # Extract only the analog channels
        self.recording_extractor = full_recording.select_channels(channel_ids=self._channels_to_include)

        super().__init__(
            file_path=self._file_path,
            channels_to_include=self._channels_to_include,
            verbose=verbose,
        )

    @property
    def channel_ids(self):
        """Gets the channel ids of the data."""
        return self.recording_extractor.get_channel_ids()

    def get_metadata(self) -> DeepDict:
        metadata = super().get_metadata()

        # Add TimeSeries metadata
        channel_names = self.channel_ids
        channels_string = ", ".join(channel_names)
        description = f"Auxiliary signals from the EDF format. Channels: {channels_string}"

        metadata["TimeSeries"] = {
            self.metadata_key: dict(
                name="TimeSeriesAnalogEDF",
                description=description,
            )
        }

        return metadata

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata: dict | None = None,
        stub_test: bool = False,
        iterator_type: str | None = "v2",
        iterator_opts: dict | None = None,
        always_write_timestamps: bool = False,
    ):
        """
        Add analog channel data to an NWB file.

        Parameters
        ----------
        nwbfile : NWBFile
            The NWB file to which the analog data will be added
        metadata : dict, optional
            Metadata dictionary with device information. If None, uses default metadata
        stub_test : bool, default: False
            If True, only writes a small amount of data for testing
        iterator_type : str, optional, default: "v2"
            Type of iterator to use for data streaming
        iterator_opts : dict, optional
            Additional options for the iterator
        always_write_timestamps : bool, default: False
            If True, always writes timestamps instead of using sampling rate
        """
        from ....tools.spikeinterface import (
            _stub_recording,
            add_recording_as_time_series_to_nwbfile,
        )

        if metadata is None:
            metadata = self.get_metadata()

        recording = self.recording_extractor
        if stub_test:
            recording = _stub_recording(recording=recording)

        add_recording_as_time_series_to_nwbfile(
            recording=recording,
            nwbfile=nwbfile,
            metadata=metadata,
            iterator_type=iterator_type,
            iterator_opts=iterator_opts,
            always_write_timestamps=always_write_timestamps,
            metadata_key=self.metadata_key,
        )



from datetime import datetime
from typing import Literal

import numpy as np
from pynwb import NWBFile
from pynwb.base import DynamicTable
from pynwb.device import Device

from .mock_ttl_signals import generate_mock_ttl_signal
from ...basedatainterface import BaseDataInterface
from ...basetemporalalignmentinterface import BaseTemporalAlignmentInterface
from ...datainterfaces import SpikeGLXNIDQInterface
from ...datainterfaces.ecephys.baserecordingextractorinterface import (
    BaseRecordingExtractorInterface,
)
from ...datainterfaces.ecephys.basesortingextractorinterface import (
    BaseSortingExtractorInterface,
)
from ...datainterfaces.ophys.baseimagingextractorinterface import (
    BaseImagingExtractorInterface,
)
from ...datainterfaces.ophys.basesegmentationextractorinterface import (
    BaseSegmentationExtractorInterface,
)
from ...tools.nwb_helpers import get_module
from ...utils import ArrayType, get_json_schema_from_method_signature
from ...utils.dict import DeepDict


class MockInterface(BaseDataInterface):
    """
    A mock interface for testing basic command passing without side effects.
    """

    def __init__(self, verbose: bool = False, **source_data):

        super().__init__(verbose=verbose, **source_data)

    def get_metadata(self) -> DeepDict:
        metadata = super().get_metadata()
        session_start_time = datetime.now().astimezone()
        metadata["NWBFile"]["session_start_time"] = session_start_time
        return metadata

    def add_to_nwbfile(self, nwbfile: NWBFile, metadata: dict | None, **conversion_options):

        return None


class MockTimeSeriesInterface(BaseDataInterface):
    """
    A mock TimeSeries interface for testing purposes.

    This interface uses pynwb's mock_TimeSeries to create synthetic time series data
    without only pynwb as a dependency.
    """

    def __init__(
        self,
        *,
        num_channels: int = 4,
        sampling_frequency: float = 30_000.0,
        duration: float = 1.0,
        seed: int = 0,
        verbose: bool = False,
        metadata_key: str = "TimeSeries",
    ):
        """
        Initialize a mock TimeSeries interface.

        Parameters
        ----------
        num_channels : int, optional
            Number of channels to generate, by default 4.
        sampling_frequency : float, optional
            Sampling frequency in Hz, by default 30,000.0 Hz.
        duration : float, optional
            Duration of the data in seconds, by default 1.0.
        seed : int, optional
            Seed for the random number generator, by default 0.
        verbose : bool, optional
            Control verbosity, by default False.
        metadata_key : str, optional
            Key for the TimeSeries metadata in the metadata dictionary, by default "TimeSeries".
        """
        self.num_channels = num_channels
        self.sampling_frequency = sampling_frequency
        self.duration = duration
        self.seed = seed
        self.metadata_key = metadata_key

        super().__init__(verbose=verbose)

    def get_metadata(self) -> DeepDict:
        """
        Get metadata for the TimeSeries interface.

        Returns
        -------
        dict
            The metadata dictionary containing NWBFile and TimeSeries metadata.
        """
        metadata = super().get_metadata()
        session_start_time = datetime.now().astimezone()
        metadata["NWBFile"]["session_start_time"] = session_start_time

        # Add TimeSeries metadata using the metadata_key
        metadata["TimeSeries"] = {
            self.metadata_key: {
                "name": self.metadata_key,
                "description": f"Mock TimeSeries data with {self.num_channels} channels",
                "unit": "n.a.",
            }
        }

        return metadata

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata: dict | None = None,
    ):
        """
        Add mock TimeSeries data to an NWB file.

        Parameters
        ----------
        nwbfile : NWBFile
            The NWB file to which the TimeSeries data will be added.
        metadata : dict, optional
            Metadata dictionary. If None, uses default metadata.
        """
        from pynwb.testing.mock.base import mock_TimeSeries

        if metadata is None:
            metadata = self.get_metadata()

        # Generate mock data
        rng = np.random.default_rng(self.seed)
        num_samples = int(self.duration * self.sampling_frequency)
        data = rng.standard_normal(size=(num_samples, self.num_channels)).astype("float32")

        # Get TimeSeries kwargs from metadata
        time_series_metadata = metadata.get("TimeSeries", {}).get(self.metadata_key, {})

        tseries_kwargs = {
            "name": time_series_metadata.get("name", "MockTimeSeries"),
            "description": time_series_metadata.get("description", "Mock TimeSeries data"),
            "unit": time_series_metadata.get("unit", "n.a."),
            "data": data,
            "starting_time": 0.0,
            "rate": self.sampling_frequency,
        }

        # Apply any additional metadata
        for key in ["comments", "conversion", "offset"]:
            if key in time_series_metadata:
                tseries_kwargs[key] = time_series_metadata[key]

        time_series = mock_TimeSeries(**tseries_kwargs)
        nwbfile.add_acquisition(time_series)


class MockBehaviorEventInterface(BaseTemporalAlignmentInterface):
    """
    A mock behavior event interface for testing purposes.
    """

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = get_json_schema_from_method_signature(method=cls.__init__, exclude=["event_times"])
        source_schema["additionalProperties"] = True
        return source_schema

    def __init__(self, event_times: ArrayType | None = None):
        """
        Initialize the interface with event times for behavior.

        Parameters
        ----------
        event_times : list of floats, optional
            The event times to set as timestamps for this interface.
            The default is the array [1.2, 2.3, 3.4] to simulate a time series similar to the
            MockSpikeGLXNIDQInterface.
        """
        event_times = event_times or [1.2, 2.3, 3.4]
        self.event_times = np.array(event_times)
        self.original_event_times = np.array(event_times)  # Make a copy of the initial loaded timestamps

    def get_original_timestamps(self) -> np.ndarray:
        """
        Get the original event times before any alignment or transformation.

        Returns
        -------
        np.ndarray
            The original event times as a NumPy array.
        """
        return self.original_event_times

    def get_timestamps(self) -> np.ndarray:
        """
        Get the current (possibly aligned) event times.

        Returns
        -------
        np.ndarray
            The current event times as a NumPy array, possibly modified after alignment.
        """
        return self.event_times

    def set_aligned_timestamps(self, aligned_timestamps: np.ndarray):
        """
        Set the event times after alignment.

        Parameters
        ----------
        aligned_timestamps : np.ndarray
            The aligned event timestamps to update the internal event times.
        """
        self.event_times = aligned_timestamps

    def add_to_nwbfile(self, nwbfile: NWBFile, metadata: dict):
        """
        Add the event times to an NWBFile as a DynamicTable.

        Parameters
        ----------
        nwbfile : NWBFile
            The NWB file to which the event times will be added.
        metadata : dict
            Metadata to describe the event times in the NWB file.

        Notes
        -----
        This method creates a DynamicTable to store event times and adds it to the NWBFile's acquisition.
        """
        table = DynamicTable(name="BehaviorEvents", description="Times of various classified behaviors.")
        table.add_column(name="event_time", description="Time of each event.")
        for timestamp in self.get_timestamps():
            table.add_row(event_time=timestamp)
        nwbfile.add_acquisition(table)


class MockSpikeGLXNIDQInterface(SpikeGLXNIDQInterface):
    """
    A mock SpikeGLX interface for testing purposes.
    """

    ExtractorName = "NumpyRecording"

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = get_json_schema_from_method_signature(method=cls.__init__, exclude=["ttl_times"])
        source_schema["additionalProperties"] = True
        return source_schema

    def __init__(
        self, signal_duration: float = 7.0, ttl_times: list[list[float]] | None = None, ttl_duration: float = 1.0
    ):
        """
        Define a mock SpikeGLXNIDQInterface by overriding the recording extractor to be a mock TTL signal.

        Parameters
        ----------
        signal_duration : float, default: 7.0
            The number of seconds to simulate.
        ttl_times : list of lists of floats, optional
            The times within the `signal_duration` to trigger the TTL pulse for each channel.
            The outer list is over channels, while each inner list is the set of TTL times for each specific channel.
            The default generates 8 channels with periodic on/off cycle (which start in the 'off' state)
            each of which is of length `ttl_duration` with a 0.1 second offset per channel.
        ttl_duration : float, default: 1.0
            How long the TTL pulses stays in the 'on' state when triggered, in seconds.
        """
        from spikeinterface.extractors import NumpyRecording

        self.has_analog_channels = True
        self.has_digital_channels = False

        if ttl_times is None:
            # Begin in 'off' state
            number_of_periods = int(np.ceil((signal_duration - ttl_duration) / (ttl_duration * 2)))
            default_periodic_ttl_times = [ttl_duration * (1 + 2 * period) for period in range(number_of_periods)]
            ttl_times = [[ttl_time + 0.1 * channel for ttl_time in default_periodic_ttl_times] for channel in range(8)]
        number_of_channels = len(ttl_times)
        channel_ids = [f"nidq#XA{channel_index}" for channel_index in range(number_of_channels)]  # NIDQ channel IDs
        channel_groups = ["NIDQChannelGroup"] * number_of_channels
        self.analog_channel_ids = channel_ids

        sampling_frequency = 25_000.0  # NIDQ sampling rate
        number_of_frames = int(signal_duration * sampling_frequency)
        traces = np.empty(shape=(number_of_frames, number_of_channels), dtype="int16")
        for channel_index in range(number_of_channels):
            traces[:, channel_index] = generate_mock_ttl_signal(
                signal_duration=signal_duration,
                ttl_times=ttl_times[channel_index],
                ttl_duration=ttl_duration,
                sampling_frequency_hz=sampling_frequency,
            )

        self.recording_extractor = NumpyRecording(
            traces_list=traces, sampling_frequency=sampling_frequency, channel_ids=channel_ids
        )
        # NIDQ channel gains
        self.recording_extractor.set_channel_gains(gains=[61.03515625] * self.recording_extractor.get_num_channels())
        self.recording_extractor.set_property(key="group_name", values=channel_groups)

        # Minimal meta so `get_metadata` works similarly to real NIDQ header
        self.meta = {"acqMnMaXaDw": "0,0,8,1", "fileCreateTime": "2020-11-03T10:35:10", "niDev1ProductName": "PCI-6259"}
        self.verbose = None
        self.es_key = "ElectricalSeriesNIDQ"


class MockRecordingInterface(BaseRecordingExtractorInterface):
    """An interface with a spikeinterface recording object for testing purposes."""

    ExtractorModuleName = "spikeinterface.core.generate"
    ExtractorName = "generate_recording"

    def __init__(
        self,
        num_channels: int = 4,
        sampling_frequency: float = 30_000.0,
        durations: tuple[float, ...] = (1.0,),
        seed: int = 0,
        verbose: bool = False,
        es_key: str = "ElectricalSeries",
        set_probe: bool = False,
    ):
        super().__init__(
            num_channels=num_channels,
            sampling_frequency=sampling_frequency,
            durations=durations,
            set_probe=set_probe,
            seed=seed,
            verbose=verbose,
            es_key=es_key,
        )

        self.recording_extractor.set_channel_gains(gains=[1.0] * self.recording_extractor.get_num_channels())
        self.recording_extractor.set_channel_offsets(offsets=[0.0] * self.recording_extractor.get_num_channels())

        # If probe was set, customize contact IDs to use "e0", "e1", etc. format for testing
        if set_probe and self.recording_extractor.has_probe():
            probe = self.recording_extractor.get_probe()
            contact_ids = [f"e{i}" for i in range(num_channels)]
            probe.set_contact_ids(contact_ids)
            self.recording_extractor = self.recording_extractor.set_probe(probe, group_mode="by_probe")

    def get_metadata(self) -> DeepDict:
        """
        Get metadata for the recording interface.

        Returns
        -------
        dict
            The metadata dictionary containing NWBFile metadata with session start time.
        """
        metadata = super().get_metadata()
        session_start_time = datetime.now().astimezone()
        metadata["NWBFile"]["session_start_time"] = session_start_time
        return metadata


class MockSortingInterface(BaseSortingExtractorInterface):
    """A mock sorting extractor interface for generating synthetic sorting data."""

    # TODO: Implement this class with the lazy generator once is merged
    # https://github.com/SpikeInterface/spikeinterface/pull/2227

    ExtractorModuleName = "spikeinterface.core.generate"
    ExtractorName = "generate_sorting"

    def __init__(
        self,
        num_units: int = 4,
        sampling_frequency: float = 30_000.0,
        durations: tuple[float, ...] = (1.0,),
        seed: int = 0,
        verbose: bool = False,
    ):
        """
        Parameters
        ----------
        num_units : int, optional
            Number of units to generate, by default 4.
        sampling_frequency : float, optional
            Sampling frequency of the generated data in Hz, by default 30,000.0 Hz.
        durations : tuple of float, optional
            Durations of the segments in seconds, by default (1.0,).
        seed : int, optional
            Seed for the random number generator, by default 0.
        verbose : bool, optional
            Control whether to display verbose messages during writing, by default True.

        """

        super().__init__(
            num_units=num_units,
            sampling_frequency=sampling_frequency,
            durations=durations,
            seed=seed,
            verbose=verbose,
        )

        # Sorting extractor to have string unit ids until is changed in SpikeInterface
        # https://github.com/SpikeInterface/spikeinterface/pull/3588
        string_unit_ids = [str(id) for id in self.sorting_extractor.unit_ids]
        self.sorting_extractor = self.sorting_extractor.rename_units(new_unit_ids=string_unit_ids)

    def get_metadata(self) -> DeepDict:
        metadata = super().get_metadata()
        session_start_time = datetime.now().astimezone()
        metadata["NWBFile"]["session_start_time"] = session_start_time
        return metadata


class MockImagingInterface(BaseImagingExtractorInterface):
    """
    A mock imaging interface for testing purposes.
    """

    ExtractorModuleName = "roiextractors.testing"
    ExtractorName = "generate_dummy_imaging_extractor"

    def __init__(
        self,
        num_samples: int | None = None,
        num_frames: int | None = None,
        num_rows: int = 10,
        num_columns: int = 10,
        sampling_frequency: float = 30,
        dtype: str = "uint16",
        verbose: bool = False,
        seed: int = 0,
        photon_series_type: Literal["OnePhotonSeries", "TwoPhotonSeries"] = "TwoPhotonSeries",
    ):
        """
        Parameters
        ----------
        num_samples : int, optional
            The number of samples (frames) in the mock imaging data, by default 30.
        num_frames : int, optional
            Deprecated. Use num_samples instead. Will be removed after February 2025.
            The number of frames in the mock imaging data, by default 30.
        num_rows : int, optional
            The number of rows (height) in each frame of the mock imaging data, by default 10.
        num_columns : int, optional
            The number of columns (width) in each frame of the mock imaging data, by default 10.
        sampling_frequency : float, optional
            The sampling frequency of the mock imaging data in Hz, by default 30.
        dtype : str, optional
            The data type of the generated imaging data (e.g., 'uint16'), by default 'uint16'.
        seed : int, optional
            Random seed for reproducibility, by default 0.
        photon_series_type : Literal["OnePhotonSeries", "TwoPhotonSeries"], optional
            The type of photon series for the mock imaging data, either "OnePhotonSeries" or
            "TwoPhotonSeries", by default "TwoPhotonSeries".
        verbose : bool, default False
            controls verbosity
        """

        # Handle deprecation of num_frames parameter
        if num_frames is not None and num_samples is not None:
            raise ValueError("Cannot specify both num_frames and num_samples. Use num_samples only.")
        elif num_frames is not None:
            import warnings

            warnings.warn(
                "The 'num_frames' parameter is deprecated and will be removed after February 2025. "
                "Use 'num_samples' instead.",
                FutureWarning,
                stacklevel=2,
            )
            num_samples = num_frames
        elif num_samples is None:
            num_samples = 30  # Default value

        self.seed = seed
        super().__init__(
            num_samples=num_samples,
            num_rows=num_rows,
            num_columns=num_columns,
            sampling_frequency=sampling_frequency,
            dtype=dtype,
            verbose=verbose,
            seed=seed,
        )

        self.verbose = verbose
        self.photon_series_type = photon_series_type

    def get_metadata(self) -> DeepDict:
        session_start_time = datetime.now().astimezone()
        metadata = super().get_metadata()
        metadata["NWBFile"]["session_start_time"] = session_start_time
        return metadata


class MockSegmentationInterface(BaseSegmentationExtractorInterface):
    """A mock segmentation interface for testing purposes."""

    ExtractorModuleName = "roiextractors.testing"
    ExtractorName = "generate_dummy_segmentation_extractor"

    def __init__(
        self,
        num_rois: int = 10,
        num_samples: int | None = None,
        num_frames: int | None = None,
        num_rows: int = 25,
        num_columns: int = 25,
        sampling_frequency: float = 30.0,
        has_summary_images: bool = True,
        has_raw_signal: bool = True,
        has_dff_signal: bool = True,
        has_deconvolved_signal: bool = True,
        has_neuropil_signal: bool = True,
        seed: int = 0,
        verbose: bool = False,
    ):
        """
        Parameters
        ----------
        num_rois : int, optional
            number of regions of interest, by default 10.
        num_samples : int, optional
            number of samples (frames), by default 30.
        num_frames : int, optional
            Deprecated. Use num_samples instead. Will be removed after February 2025.
            number of frames, by default 30.
        num_rows : int, optional
            number of rows in the hypothetical video from which the data was extracted, by default 25.
        num_columns : int, optional
            number of columns in the hypothetical video from which the data was extracted, by default 25.
        sampling_frequency : float, optional
            sampling frequency of the hypothetical video from which the data was extracted, by default 30.0.
        has_summary_images : bool, optional
            whether the dummy segmentation extractor has summary images or not (mean and correlation).
        has_raw_signal : bool, optional
            whether a raw fluorescence signal is desired in the object, by default True.
        has_dff_signal : bool, optional
            whether a relative (df/f) fluorescence signal is desired in the object, by default True.
        has_deconvolved_signal : bool, optional
            whether a deconvolved signal is desired in the object, by default True.
        has_neuropil_signal : bool, optional
            whether a neuropil signal is desired in the object, by default True.
        seed: int, default 0
            seed for the random number generator, by default 0
        verbose : bool, optional
            controls verbosity, by default False.
        """

        # Handle deprecation of num_frames parameter
        if num_frames is not None and num_samples is not None:
            raise ValueError("Cannot specify both num_frames and num_samples. Use num_samples only.")
        elif num_frames is not None:
            import warnings

            warnings.warn(
                "The 'num_frames' parameter is deprecated and will be removed after February 2025. "
                "Use 'num_samples' instead.",
                FutureWarning,
                stacklevel=2,
            )
            num_samples = num_frames
        elif num_samples is None:
            num_samples = 30  # Default value

        super().__init__(
            num_rois=num_rois,
            num_samples=num_samples,
            num_rows=num_rows,
            num_columns=num_columns,
            sampling_frequency=sampling_frequency,
            has_summary_images=has_summary_images,
            has_raw_signal=has_raw_signal,
            has_dff_signal=has_dff_signal,
            has_deconvolved_signal=has_deconvolved_signal,
            has_neuropil_signal=has_neuropil_signal,
            verbose=verbose,
            seed=seed,
        )

    def get_metadata(self) -> DeepDict:
        session_start_time = datetime.now().astimezone()
        metadata = super().get_metadata()
        metadata["NWBFile"]["session_start_time"] = session_start_time
        return metadata


class MockPoseEstimationInterface(BaseTemporalAlignmentInterface):
    """
    A mock pose estimation interface for testing purposes.
    """

    display_name = "Mock Pose Estimation"
    keywords = (
        "behavior",
        "pose estimation",
        "mock",
    )
    associated_suffixes = []
    info = "Mock interface for pose estimation data testing."

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = get_json_schema_from_method_signature(method=cls.__init__, exclude=["timestamps", "confidence"])
        source_schema["additionalProperties"] = True
        return source_schema

    def __init__(
        self,
        num_samples: int = 1000,
        num_nodes: int = 3,
        pose_estimation_metadata_key: str = "MockPoseEstimation",
        seed: int = 0,
        verbose: bool = False,
    ):
        """
        Initialize a mock pose estimation interface.

        Parameters
        ----------
        num_samples : int, optional
            Number of samples to generate, by default 1000.
        num_nodes : int, optional
            Number of nodes/body parts to track, by default 3.
        pose_estimation_metadata_key : str, optional
            Key for pose estimation metadata container, by default "MockPoseEstimation".
        seed : int, optional
            Random seed for reproducible data generation, by default 0.
        verbose : bool, optional
            Control verbosity, by default False.
        """
        self.num_samples = num_samples
        self.num_nodes = num_nodes
        self.pose_estimation_metadata_key = pose_estimation_metadata_key
        self.seed = seed
        self.verbose = verbose

        # Set metadata defaults
        self.scorer = "MockScorer"
        self.source_software = "MockSourceSoftware"

        # Generate random nodes and edges
        orbital_body_parts = [
            "head",
            "neck",
            "left_shoulder",
            "right_shoulder",
            "chest",
            "left_elbow",
            "right_elbow",
            "left_wrist",
            "right_wrist",
            "pelvis",
        ]

        # Use orbital body parts if we have enough, otherwise generate generic nodes
        if num_nodes <= len(orbital_body_parts):
            self.nodes = orbital_body_parts[:num_nodes]
        else:
            self.nodes = orbital_body_parts + [f"node_{i}" for i in range(len(orbital_body_parts), num_nodes)]

        # Generate random edges (connect some nodes randomly)
        np.random.seed(seed)  # For reproducible edge generation
        num_edges = min(num_nodes - 1, max(1, num_nodes // 2))  # Reasonable number of edges
        possible_edges = [(i, j) for i in range(num_nodes) for j in range(i + 1, num_nodes)]
        selected_edges = np.random.choice(len(possible_edges), size=num_edges, replace=False)
        self.edges = np.array([possible_edges[i] for i in selected_edges], dtype="uint8")

        # Generate timestamps (private attributes)
        self._original_timestamps = np.linspace(0.0, float(num_samples) / 30.0, num_samples)
        self._timestamps = np.copy(self._original_timestamps)

        # Generate pose estimation data
        self.pose_data = self._generate_pose_data()

        super().__init__(verbose=verbose)

        # Import ndx_pose to ensure it's available
        import ndx_pose  # noqa: F401

    def _generate_pose_data(self) -> np.ndarray:
        """Generate pose estimation data with center following Lissajous trajectory and nodes fixed on circle."""
        # Fixed to 2D for now
        shape = (self.num_samples, self.num_nodes, 2)

        # Generate Lissajous trajectory for the center
        time_points = np.linspace(0, 4 * np.pi, self.num_samples)
        center_x = 320 + 80 * np.sin(1.2 * time_points)  # Center follows Lissajous
        center_y = 240 + 60 * np.sin(1.7 * time_points + np.pi / 3)

        # Generate data for all nodes
        data = np.zeros(shape)
        circle_radius = 50  # Radius of circle around center

        for node_index in range(self.num_nodes):
            # Position each node equally spaced around a circle relative to center
            angle = 2 * np.pi * node_index / self.num_nodes

            # Fixed position on circle relative to center (no oscillations)
            offset_x = circle_radius * np.cos(angle)
            offset_y = circle_radius * np.sin(angle)

            # Final position: center + fixed circle position
            data[:, node_index, 0] = center_x + offset_x
            data[:, node_index, 1] = center_y + offset_y

        return data

    def get_original_timestamps(self) -> np.ndarray:
        """Get the original timestamps before any alignment."""
        return self._original_timestamps

    def get_timestamps(self) -> np.ndarray:
        """Get the current (possibly aligned) timestamps."""
        return self._timestamps

    def set_aligned_timestamps(self, aligned_timestamps: np.ndarray):
        """Set aligned timestamps."""
        self._timestamps = aligned_timestamps

    def get_metadata(self) -> DeepDict:
        """Get metadata for the mock pose estimation interface."""
        metadata = super().get_metadata()
        session_start_time = datetime.now().astimezone()
        metadata["NWBFile"]["session_start_time"] = session_start_time

        # Create metadata following the DeepLabCut pattern
        container_name = self.pose_estimation_metadata_key
        skeleton_name = f"Skeleton{container_name}"
        device_name = f"Camera{container_name}"

        # Create PoseEstimation metadata structure
        pose_estimation_metadata = DeepDict()

        # Add Skeleton as a dictionary
        pose_estimation_metadata["Skeletons"] = {
            skeleton_name: {"name": skeleton_name, "nodes": self.nodes, "edges": self.edges.tolist()}
        }

        # Add Device as a dictionary
        pose_estimation_metadata["Devices"] = {
            device_name: {"name": device_name, "description": "Mock camera device for pose estimation testing."}
        }

        # Add PoseEstimation container
        pose_estimation_metadata["PoseEstimationContainers"] = {
            container_name: {
                "name": container_name,
                "description": f"Mock pose estimation data from {self.source_software}.",
                "source_software": self.source_software,
                "dimensions": [[640, 480]],
                "skeleton": skeleton_name,
                "devices": [device_name],
                "scorer": self.scorer,
                "original_videos": ["mock_video.mp4"],
                "PoseEstimationSeries": {},
            }
        }

        # Add a series for each node
        for node in self.nodes:
            # Convert node name to PascalCase for the series name
            pascal_case_node = "".join(word.capitalize() for word in node.replace("_", " ").split())
            series_name = f"PoseEstimationSeries{pascal_case_node}"

            pose_estimation_metadata["PoseEstimationContainers"][container_name]["PoseEstimationSeries"][node] = {
                "name": series_name,
                "description": f"Mock pose estimation series for {node}.",
                "unit": "pixels",
                "reference_frame": "(0,0) corresponds to the bottom left corner of the video.",
                "confidence_definition": "Softmax output of the deep neural network.",
            }

        # Add PoseEstimation metadata to the main metadata
        metadata["PoseEstimation"] = pose_estimation_metadata

        return metadata

    def add_to_nwbfile(self, nwbfile: NWBFile, metadata: dict | None = None, **conversion_options):
        """Add mock pose estimation data to NWBFile using ndx-pose."""
        from ndx_pose import PoseEstimation, PoseEstimationSeries, Skeleton, Skeletons

        # Create or get behavior processing module
        behavior_module = get_module(nwbfile, "behavior")

        # Create device
        device = Device(name="MockCamera", description="Mock camera device for pose estimation testing")
        nwbfile.add_device(device)

        # Create skeleton
        skeleton = Skeleton(name="MockSkeleton", nodes=self.nodes, edges=self.edges)

        # Create pose estimation series for each node
        pose_estimation_series = []
        for index, node_name in enumerate(self.nodes):
            # Convert node name to PascalCase for the series name
            pascal_case_node = "".join(word.capitalize() for word in node_name.replace("_", " ").split())
            series_name = f"PoseEstimationSeries{pascal_case_node}"

            series = PoseEstimationSeries(
                name=series_name,
                description=f"Pose estimation for {node_name}",
                data=self.pose_data[:, index, :],
                unit="pixels",
                reference_frame="top left corner of video frame",
                timestamps=self.get_timestamps(),
                confidence=np.ones(self.num_samples),
                confidence_definition="definition of confidence",
            )
            pose_estimation_series.append(series)

        # Create pose estimation container
        pose_estimation = PoseEstimation(
            name="MockPoseEstimation",
            description=f"Mock pose estimation data from {self.source_software}",
            pose_estimation_series=pose_estimation_series,
            skeleton=skeleton,
            devices=[device],
            scorer=self.scorer,
            source_software=self.source_software,
            dimensions=np.array([[640, 480]], dtype="uint16"),
            original_videos=["mock_video.mp4"],
            labeled_videos=["mock_video_labeled.mp4"],
        )

        behavior_module.add(pose_estimation)
        if "Skeletons" not in behavior_module.data_interfaces:
            skeletons = Skeletons(skeletons=[skeleton])
            behavior_module.add(skeletons)
        else:
            skeletons = behavior_module["Skeletons"]
            skeletons.add_skeletons(skeleton)
